import re
import logging
from django.core.cache import get_cache
from django.db import IntegrityError
from vdw.pipeline import PIPELINE_CACHE_ALIAS
from vdw.genome.models import Chromosome
from vdw.variants.models import Variant, VariantType, Effect, \
    FunctionalClass, VariantEffect
from vdw.variants.utils import calculate_md5
from vdw.raw.utils.stream import VCFPGCopyEditor
from vdw.genes.models import Gene, Transcript
from vdw.pipeline import checks

log = logging.getLogger(__name__)


class VariantCache(object):
    """Variant cache and lookup by chromosome and position.

    If `shared` cache is enabled, this enables parallelization since this
    cache ensures atomic gets and sets. The shared cache acts as central
    reference for processes to check against.

    For loading, if a record is present in the cache, it has already been
    processed and thus can be skipped by other processes attempting to load the
    variant.

    The other use case is a central cache for looking up variant keys. This
    prevents having in-process caches that possibly redundant.
    """
    def __init__(self, size=1000, shared=True, timeout=60 * 60 * 24):
        self.size = size
        self.chrom = None
        self.start = None
        self.end = None

        self.shared = shared
        if shared:
            # Only applies to the shared cache..
            self.timeout = timeout
            self._cache = get_cache(PIPELINE_CACHE_ALIAS)
        else:
            self._cache = {}

    def _get_variants(self):
        return Variant.objects.filter(
            chr__value=self.chrom,
            pos__range=(self.start, self.end)).values_list('md5', 'pk')

    def ensure_cache(self, record):
        if (record.CHROM != self.chrom or
                not (self.start <= record.POS <= self.end)):
            self.chrom = record.CHROM
            self.start = record.POS
            self.end = self.start + self.size

            if self.shared:
                return

            self._cache = dict(self._get_variants())

    def __contains__(self, md5):
        if not self.shared:
            return md5 in self._cache

        # If the add is successful, that means it is not present in the
        # shared cached and thus should return False which will cause the
        # variant to be loaded. Note, this acts as a placeholder and does
        # not store the primary key of variant. This is generally not an
        # issue since subsequent processes will always load a fresh
        # snapshot of variants from the database.
        if not self._cache.add(md5, -1, self.timeout):
            return True

        present = False

        # Since these are lazily cached, hit the database for a slice of
        # variants.
        variants = self._get_variants()
        for _md5, _pk in variants.iterator():
            if _md5 == md5:
                present = True
            self._cache.set(_md5, _pk, self.timeout)

        return present

    def get(self, md5, default=None):
        if not self.shared:
            return self._cache.get(md5, default)

        pk = self._cache.get(md5)

        # Ensure this is not a placholder
        if pk is None or pk == -1:
            variants = self._get_variants()
            for _md5, _pk in variants.iterator():
                if _md5 == md5:
                    pk = _pk
                self._cache.set(_md5, _pk, self.timeout)
        return pk


class VariantStream(VCFPGCopyEditor):
    """
    Reads a VCF file and loads variants that are not already in the database.

    An instance of `VariantCache` is used to determine whether a variant is
    present in the database or not.
    """
    vcf_fields = ('CHROM', 'POS', 'REF', 'ALT', 'ID')

    output_columns = ('chr_id', 'pos', 'ref', 'alt', 'rsid', 'type_id', 'md5')

    def __init__(self, *args, **kwargs):
        self.use_cache = kwargs.pop('use_cache', True)
        cache_size = kwargs.pop('cache_size', 1000)
        super(VariantStream, self).__init__(*args, **kwargs)

        # Cache of chromosomes and their keys
        self.chromosomes = \
            dict([(x.value, x) for x in Chromosome.objects.all()])

        # Cache variant types and their keys
        self.variant_types = \
            dict([(x.value, x) for x in VariantType.objects.all()])

        # Special cache for variants
        self.variants = VariantCache(cache_size)

    def process_line(self, record):
        cleaned = super(VariantStream, self).process_line(record)

        # Replace chromosome with primary key
        cleaned[0] = str(self.chromosomes[cleaned[0]].pk)

        # Ensure the variant type exists.. this will *rarely* get updated
        var_type = record.var_type.upper()
        if var_type not in self.variant_types:
            vt = VariantType(label=var_type, value=var_type)
            try:
                vt.save()
            except IntegrityError:
                vt = VariantType.objects.get(label=var_type, value=var_type)
            self.variant_types[var_type] = vt

        # Add variant type
        cleaned.append(str(self.variant_types[var_type].pk))

        return cleaned

    def readline(self, size=-1):
        "Ignore the `size` since a complete line must be processed."
        while True:
            try:
                record = next(self.reader)
            except StopIteration:
                break

            # Ensure this is a valid record
            if checks.record_is_valid(record):
                if self.use_cache:
                    # Ensures the cache is updated and available
                    self.variants.ensure_cache(record)

                # Calculate the MD5 of the variant itself (not the record)
                md5 = calculate_md5(record)

                # Ensure this variant is not already loaded
                if not self.use_cache or md5 not in self.variants:
                    cleaned = self.process_line(record)
                    cleaned.append(md5)
                    return self.outdel.join(cleaned) + '\n'

        return ''


class EffectStream(VCFPGCopyEditor):
    vcf_fields = ('CHROM', 'POS', 'REF', 'ALT')

    output_columns = ('variant_id', 'codon_change', 'amino_acid_change',
                      'effect_id', 'functional_class_id', 'gene_id',
                      'transcript_id', 'segment', 'hgvs_c', 'hgvs_p')

    def __init__(self, *args, **kwargs):
        self.chromosomes = \
            dict([(x.value, x) for x in Chromosome.objects.all()])
        self.effects = dict(Effect.objects.values_list('value', 'pk'))
        self.functional_classes = \
            dict(FunctionalClass.objects.values_list('value', 'pk'))
        self.genes = dict(Gene.objects.values_list('symbol', 'id'))
        self.transcripts = \
            dict(Transcript.objects.values_list('refseq_id', 'id'))
        self.variants = VariantCache()

        self.skip_existing = kwargs.pop('skip_existing', True)
        self._cache = get_cache(PIPELINE_CACHE_ALIAS)
        self._cache_timeout = 60 * 60 * 24

        super(EffectStream, self).__init__(*args, **kwargs)

        # This is pretty hideous, but to prevent added/removed snpEff fields
        # breaking the code, the labels needs to be mapped to the fields by
        # position. The actual `process_line` method will use the dict to
        # access snpEff values.
        self._snpeff_regex = re.compile(r'(\w+)\s*\((.*)\)')

        # Remove optional fields
        if 'EFF' in self.reader.infos:
            desc = self.reader.infos['EFF'].desc.replace(
                '[ | ERRORS | WARNINGS ]', '')

            # Get available fields, a dict will be created in process_line
            self.snpeff_keys = self._parse_snpeff_row(desc)[1]

    def _parse_snpeff_row(self, row):
        effect, tail = self._snpeff_regex.search(row).groups()
        effect = effect.strip()
        tail = [x.strip() for x in tail.split('|')]
        return effect, tail

    def _snpeff_dict(self, row):
        effect, values = self._parse_snpeff_row(row)
        return effect, dict(zip(self.snpeff_keys, values))

    def get_gene(self, gene_name):
        """Get a gene from the cache or attempt to disambiguate or add a
        new record.
        """
        if not gene_name:
            return

        gene_pk = self.genes.get(gene_name)
        if gene_pk:
            return gene_pk

        chrom = self.chromosomes[self.variants.chrom]
        gene = Gene.objects.find(gene_name, chrom, create=True)
        if gene:
            return gene.pk

    def get_transcript(self, gene_pk, refseq_id):
        "Get a transcript from the cache or add a new record."
        if not refseq_id:
            return
        transcript_pk = self.transcripts.get(refseq_id)
        if transcript_pk:
            return transcript_pk
        gene = Gene(pk=gene_pk)
        transcript = Transcript(refseq_id=refseq_id, gene=gene)
        try:
            transcript.save()
        except IntegrityError:
            transcript = Transcript.objects.get(refseq_id=refseq_id, gene=gene)
        self.transcripts[refseq_id] = transcript.pk
        return transcript.pk

    def _effects_exists(self, md5, variant_id):
        if not self.skip_existing:
            return False

        key = '{0}:effects'.format(md5)
        # If a successful add occurred, this means it is not in the cache,
        # fallback to database check
        if self._cache.add(key, 1, self._cache_timeout):
            return VariantEffect.objects.filter(
                variant__pk=variant_id).exists()
        return True

    def process_line(self, record):
        # Calculate MD5 using extracted values
        md5 = calculate_md5(record)

        # Ensures the cache is updated and available
        self.variants.ensure_cache(record)

        # Ensure the variant exists
        variant_id = self.variants.get(md5)

        if not variant_id:
            log.error('Missing variant', extra={
                'chr': record.CHROM,
                'pos': record.POS,
                'ref': record.REF,
                'alt': record.ALT,
            })
            return

        # Skip processing effects since they are only loaded once
        if self._effects_exists(md5, variant_id):
            return

        effects = []

        # Is this returning a list now??
        effects_line = record.INFO.get('EFF', [])

        # Multiple separate SNPEff records
        for eff in effects_line:
            effect, values = self._snpeff_dict(eff)

            transcript = values.get('Transcript')
            gene_pk = self.get_gene(values.get('Gene_Name'))

            row = [
                variant_id,
                values.get('Codon_Change'),
                values.get('Amino_acid_change'),
                self.effects.get(effect),
                self.functional_classes.get(values.get('Functional_Class')),
                gene_pk,
                self.get_transcript(gene_pk, transcript),
            ]

            # Extension fields from snpEff CBMi fork
            segment = values.get('Segment')
            hgvs_c = values.get('HGVS_DNA_nomenclature')
            hgvs_p = values.get('HGVS_protein_nomenclature')

            # Trim off transcript prefix, clean up format
            if transcript:
                if segment and segment.startswith(transcript):
                    segment = \
                        segment[len(transcript):].strip('._').replace('_', '.')
                if hgvs_c and hgvs_c.startswith(transcript):
                    hgvs_c = hgvs_c[len(transcript):].lstrip(':')

            row.extend([
                segment,
                hgvs_c,
                hgvs_p,
            ])

            effects.append([self.process_column('', x) for x in row])

        if not effects:
            log.error('No effects process for variant', extra={
                'chr': record.CHROM,
                'pos': record.POS,
                'ref': record.REF,
                'alt': record.ALT,
            })

        return effects

    def readline(self, size=-1):
        "Ignore the `size` since a complete line must be processed."
        while True:
            try:
                record = next(self.reader)
            except StopIteration:
                break

            # Ensure this is a valid record
            if checks.record_is_valid(record):
                # Process the line
                cleaned = self.process_line(record)
                if cleaned:
                    lines = []
                    for c in cleaned:
                        lines.append(self.outdel.join([str(x) for x in c]))
                    return '\n'.join(lines) + '\n'

        return ''



from django.db import models
from avocado.lexicon.models import Lexicon


class Chromosome(Lexicon):
    label = models.CharField(max_length=2)
    value = models.CharField(max_length=2, db_index=True)

    class Meta(Lexicon.Meta):
        db_table = 'chromosome'


class Genotype(Lexicon):
    label = models.CharField(max_length=20)
    value = models.CharField(max_length=3)

    class Meta(object):
        db_table = 'genotype'



from vdw.management.commands import Subcommander


class Command(Subcommander):
    app_name = 'raw'

    subcommands = {
        'evs': 'evs',
        'hgnc': 'hgnc',
        'polyphen2': 'polyphen2',
        'sift': 'sift',
        '1000g': 'thousandg',
        'ucsc': 'ucsc',
        'hpo': 'hpo',
    }

    import_template = 'vdw.{app}.{module}.management.subcommands.load'



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):

        # Removing unique constraint on 'CohortVariant', fields ['variant', 'cohort']
        db.delete_unique('cohort_variant', ['variant_id', 'cohort_id'])

        # Removing unique constraint on 'Cohort', fields ['project', 'name']
        db.delete_unique('project_cohort', ['project_id', 'name'])

        # Removing unique constraint on 'Sample', fields ['cohort', 'name']
        db.delete_unique('sample', ['cohort_id', 'name'])

        # Deleting model 'CohortVariant'
        db.delete_table('cohort_variant')

        # Renaming model 'Cohort' to 'Batch'
        db.rename_table('project_cohort', 'batch')

        # Adding unique constraint on 'Batch', fields ['project', 'name']
        db.create_unique('batch', ['project_id', 'name'])

        # Changing field 'Sample.cohort'
        db.alter_column('sample', 'cohort_id', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['samples.Batch']))

        db.rename_column('sample', 'cohort_id', 'batch_id')

        # Adding unique constraint on 'Sample', fields ['batch', 'name']
        db.create_unique('sample', ['batch_id', 'name'])

    def backwards(self, orm):

        # Removing unique constraint on 'Batch', fields ['project', 'name']
        db.delete_unique('batch', ['project_id', 'name'])

        db.rename_table('batch', 'project_cohort')

        db.rename_column('sample', 'batch_id', 'cohort_id')

        # Changing field 'Sample.cohort'
        db.alter_column('sample', 'cohort_id', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['samples.Cohort']))

        # Adding unique constraint on 'Cohort', fields ['project', 'name']
        db.create_unique('project_cohort', ['project_id', 'name'])

        # Adding model 'CohortVariant'
        db.create_table('cohort_variant', (
            ('cohort', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['samples.Cohort'])),
            ('variant', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['variants.Variant'])),
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('af', self.gf('django.db.models.fields.FloatField')(null=True, db_index=True)),
        ))
        db.send_create_signal('samples', ['CohortVariant'])

        # Adding unique constraint on 'CohortVariant', fields ['variant', 'cohort']
        db.create_unique('cohort_variant', ['variant_id', 'cohort_id'])


    models = {
        'genome.chromosome': {
            'Meta': {'ordering': "['order']", 'object_name': 'Chromosome', 'db_table': "'chromosome'"},
            'code': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '2'}),
            'order': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'value': ('django.db.models.fields.CharField', [], {'max_length': '2', 'db_index': 'True'})
        },
        'genome.genome': {
            'Meta': {'object_name': 'Genome', 'db_table': "'genome'"},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            'released': ('django.db.models.fields.DateField', [], {'null': 'True'}),
            'version': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'genome.genotype': {
            'Meta': {'object_name': 'Genotype', 'db_table': "'genotype'"},
            'code': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '20'}),
            'order': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'value': ('django.db.models.fields.CharField', [], {'max_length': '3'})
        },
        'literature.pubmed': {
            'Meta': {'object_name': 'PubMed', 'db_table': "'pubmed'"},
            'pmid': ('django.db.models.fields.IntegerField', [], {'primary_key': 'True'})
        },
        'phenotypes.phenotype': {
            'Meta': {'object_name': 'Phenotype', 'db_table': "'phenotype'"},
            'articles': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['literature.PubMed']", 'symmetrical': 'False'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'hpo_id': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'term': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '1000'})
        },
        'samples.batch': {
            'Meta': {'unique_together': "(('project', 'name'),)", 'object_name': 'Batch', 'db_table': "'batch'"},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'investigator': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'project': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['samples.Project']"}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'})
        },
        'samples.person': {
            'Meta': {'object_name': 'Person', 'db_table': "'person'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'mrn': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'proband': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'relations': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['samples.Person']", 'through': "orm['samples.Relation']", 'symmetrical': 'False'}),
            'sex': ('django.db.models.fields.CharField', [], {'max_length': '20', 'null': 'True', 'blank': 'True'})
        },
        'samples.project': {
            'Meta': {'unique_together': "(('name',),)", 'object_name': 'Project', 'db_table': "'project'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        'samples.relation': {
            'Meta': {'ordering': "('person', '-generation')", 'object_name': 'Relation', 'db_table': "'relation'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'generation': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'person': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'family'", 'to': "orm['samples.Person']"}),
            'relative': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'relative_of'", 'to': "orm['samples.Person']"}),
            'type': ('django.db.models.fields.CharField', [], {'max_length': '20'})
        },
        'samples.result': {
            'Meta': {'unique_together': "(('sample', 'variant'),)", 'object_name': 'Result', 'db_table': "'sample_result'"},
            'baseq_rank_sum': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'coverage_alt': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'coverage_ref': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'downsampling': ('django.db.models.fields.NullBooleanField', [], {'null': 'True', 'blank': 'True'}),
            'fisher_strand': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'genotype': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['genome.Genotype']", 'null': 'True', 'blank': 'True'}),
            'genotype_quality': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'haplotype_score': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'homopolymer_run': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'in_dbsnp': ('django.db.models.fields.NullBooleanField', [], {'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'mq': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'mq0': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'mq_rank_sum': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'phred_scaled_likelihood': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'quality': ('django.db.models.fields.FloatField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'quality_by_depth': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'read_depth': ('django.db.models.fields.IntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'read_pos_rank_sum': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'sample': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['samples.Sample']"}),
            'spanning_deletions': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'strand_bias': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'variant': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['variants.Variant']"})
        },
        'samples.sample': {
            'Meta': {'unique_together': "(('batch', 'name'),)", 'object_name': 'Sample', 'db_table': "'sample'"},
            'bio_sample': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'batch': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'samples'", 'to': "orm['samples.Batch']"}),
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'md5': ('django.db.models.fields.CharField', [], {'max_length': '32', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'person': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'samples'", 'null': 'True', 'to': "orm['samples.Person']"}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'version': ('django.db.models.fields.IntegerField', [], {})
        },
        'samples.samplerun': {
            'Meta': {'object_name': 'SampleRun', 'db_table': "'sample_run'"},
            'completed': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'genome': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['genome.Genome']", 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'sample': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['samples.Sample']"})
        },
        'variants.variant': {
            'Meta': {'unique_together': "(('chr', 'pos', 'ref', 'alt'),)", 'object_name': 'Variant', 'db_table': "'variant'"},
            'alt': ('django.db.models.fields.TextField', [], {'db_index': 'True'}),
            'articles': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['literature.PubMed']", 'db_table': "'variant_pubmed'", 'symmetrical': 'False'}),
            'chr': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['genome.Chromosome']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'liftover': ('django.db.models.fields.NullBooleanField', [], {'null': 'True', 'blank': 'True'}),
            'md5': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'phenotypes': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['phenotypes.Phenotype']", 'through': "orm['variants.VariantPhenotype']", 'symmetrical': 'False'}),
            'pos': ('django.db.models.fields.IntegerField', [], {}),
            'ref': ('django.db.models.fields.TextField', [], {'db_index': 'True'}),
            'rsid': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['variants.VariantType']", 'null': 'True'})
        },
        'variants.variantphenotype': {
            'Meta': {'object_name': 'VariantPhenotype', 'db_table': "'variant_phenotype'"},
            'hgmd_id': ('django.db.models.fields.CharField', [], {'max_length': '30', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'phenotype': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['phenotypes.Phenotype']"}),
            'variant': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['variants.Variant']"})
        },
        'variants.varianttype': {
            'Meta': {'ordering': "['order']", 'object_name': 'VariantType', 'db_table': "'variant_type'"},
            'code': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '20'}),
            'order': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'value': ('django.db.models.fields.CharField', [], {'max_length': '20'})
        }
    }

    complete_apps = ['samples']



import re
from vdw.raw.management.subcommands.load import LoadCommand
from vdw.raw.utils.stream import PGCopyEditor

valid_date_re = re.compile('\d{4}-\d{2}-\d{2}')


class HGNCGeneParser(PGCopyEditor):
    def process_column(self, idx, value):
        if not value:
            value = None
        elif idx == 0:
            return value[5:]
        elif idx in (19, 38, 39):
            return value[4:]
        elif idx in (11, 12, 13, 14):
            if valid_date_re.match(value):
                return value
        return super(HGNCGeneParser, self).process_column(idx, value)


class HGNCGeneFamilyParser(PGCopyEditor):
    def __init__(self, *args, **kwargs):
        kwargs['indices'] = [1, 2, 3, 4]
        super(HGNCGeneFamilyParser, self).__init__(*args, **kwargs)


class Command(LoadCommand):
    targets = ['hgnc', 'hgncFamilies']

    qualified_names = {
        'hgnc': '"raw"."hgnc"',
        'hgncFamilies': '"raw"."hgnc_families"',
    }

    drop_sql = {
        'hgnc': 'DROP TABLE IF EXISTS {}',
        'hgncFamilies': 'DROP TABLE IF EXISTS {}',
    }

    create_sql = {
        'hgnc': '''CREATE TABLE {} (
            "hgnc_id" integer NOT NULL PRIMARY KEY,
            "approved_symbol" varchar(255) NOT NULL,
            "approved_name" text NOT NULL,
            "status" varchar(50) NOT NULL,
            "locus_type" varchar(100) NOT NULL,
            "locus_group" varchar(100) NOT NULL,
            "previous_symbols" text,
            "previous_names" text,
            "synonyms" text,
            "name_synonyms" text,
            "chromosome" varchar(255),
            "date_approved" date,
            "date_modified" date,
            "date_symbol_changed" date,
            "date_name_changed" date,
            "accession_numbers" text,
            "enzyme_ids" text,
            "entrez_gene_id" integer,
            "ensembl_gene_id" varchar(50),
            "mgi_id" varchar(50),
            "specialist_database_links" text,
            "specialist_database_ids" text,
            "pubmed_ids" text,
            "refseq_ids" text,
            "gene_family_tag" text,
            "gene_family_description" text,
            "record_type" varchar(50),
            "primary_ids" text,
            "secondary_ids" text,
            "ccds_ids" text,
            "vega_ids" text,
            "locus_specified_databases" text,
            "ncbi_entrez_gene_id" integer,
            "ncbi_omim_id" integer,
            "ncbi_refseq_id" varchar(50),
            "ncbi_uniprot_id" varchar(50),
            "ncbi_ensembl_id" varchar(50),
            "ncbi_ucsc_id" varchar(50),
            "ncbi_mgi_id" integer,
            "ncbi_rgd_id" integer
        )''',

        'hgncFamilies': '''CREATE TABLE {} (
            "tag" varchar(20) NOT NULL,
            "description" varchar(200) NOT NULL,
            "symbol" varchar(20) NOT NULL,
            "hgnc_id" int NOT NULL
        )'''
    }

    processors = {
        'hgnc': HGNCGeneParser,
        'hgncFamilies': HGNCGeneFamilyParser,
    }



import logging
import ConfigParser
from functools import wraps
from django.dispatch import Signal
from rq.decorators import job as _rq_job
from django_rq.queues import get_queue

log = logging.getLogger(__name__)


class ManifestReader(object):
    def __init__(self, path):
        self.path = path
        self.config = ConfigParser.ConfigParser()
        try:
            self.config.read([path])
        except ConfigParser.MissingSectionHeaderError:
            log.warning('Empty manifest file', extra={
                'path': self.path,
            })

    def option(self, section, option, default=None):
        if self.config.has_option(section, option):
            return self.config.get(section, option)
        log.warning('No manifest option: {0}.{1}'.format(section, option),
                    extra={'path': self.path})
        return default

    def has_section(self, section):
        return self.config.has_section(section)

    def section(self, section):
        if self.config.has_section(section):
            return dict(self.config.items(section))
        log.warning('No manifest section: {0}'.format(section), extra={
            'path': self.path,
        })
        return {}

    def marked_for_load(self):
        load = self.option('general', 'load')
        return load is not None and load.lower() == 'true'


class _job(_rq_job):
    def __call__(self, f):
        super(_job, self).__call__(f)

        # Function that acts as the receiver for channels
        @wraps(f)
        def receive(signal, sender, **kwargs):
            f.delay(**kwargs)

        @wraps(f)
        def watch(channel):
            channel.subscribe(receive)

        f.receive = receive
        f.watch = watch

        return f


# Taken from django-rq code base to utilize custom _job decorator class
def job(func_or_queue=None, connection=None, *args, **kwargs):
    """
    The same as RQ's job decorator, but it works automatically works out
    the ``connection`` argument from RQ_QUEUES.

    And also, it allows simplified ``@job`` syntax to put job into
    default queue.
    """
    if callable(func_or_queue):
        func = func_or_queue
        queue = 'default'
    else:
        func = None
        queue = func_or_queue or 'default'

    if not isinstance(queue, basestring):
        queue = unicode(queue)

    try:
        queue = get_queue(queue)
        if connection is None:
            connection = queue.connection
    except KeyError:
        pass

    decorator = _job(queue, connection=connection, *args, **kwargs)
    if func:
        return decorator(func)
    return decorator


class Channel(object):
    def __init__(self, name, providing_args):
        self.name = name
        self.signal = Signal(providing_args=providing_args)

    def __str__(self):
        return str(self.name)

    def __unicode__(self):
        return unicode(self.name)

    def __repr__(self):
        return '<Channel: {0}>'.format(self.name)

    def publish(self, **kwargs):
        "Publishes to the channel which notifies all connected handlers."
        log.debug('Publish to {0}'.format(self))
        self.signal.send(sender=self.name, **kwargs)

    def subscribe(self, receiver):
        "Subscribes an external handler to this channel."
        log.debug('{0}.{1} subscribe to {2}'
                  .format(receiver.__module__, receiver.__name__, self))
        self.signal.connect(receiver)



import re
from vdw.raw.management.subcommands.load import LoadCommand
from vdw.raw.utils.stream import PGCopyEditor

ccds_exon_id_re = re.compile(r'([\w\d]+\.\d+)_exon_(\d+).*')


class CCDSExonsParser(PGCopyEditor):
    def process_column(self, idx, value):
        # parse CCDS ID and exon number
        if idx == 3:
            return ccds_exon_id_re.search(value).groups()
        return super(CCDSExonsParser, self).process_column(idx, value)


class Command(LoadCommand):

    targets = ['knownGene', 'kgAlias', 'refGene', 'ccdsKgMap', 'ccdsExons',
               'knownToRefSeq']

    qualified_names = {
        'knownGene': '"raw"."ucsc_known_gene"',
        'kgAlias': '"raw"."ucsc_kg_alias"',
        'refGene': '"raw"."ucsc_ref_gene"',
        'ccdsKgMap': '"raw"."ucsc_ccds_kg_map"',
        'ccdsExons': '"raw"."ucsc_ccds_exons"',
        'knownToRefSeq': '"raw"."ucsc_known_to_refseq"',
    }

    drop_sql = {
        'knownGene': 'DROP TABLE IF EXISTS {}',
        'kgAlias': 'DROP TABLE IF EXISTS {}',
        'refGene': 'DROP TABLE IF EXISTS {}',
        'ccdsKgMap': 'DROP TABLE IF EXISTS {}',
        'ccdsExons': 'DROP TABLE IF EXISTS {}',
        'knownToRefSeq': 'DROP TABLE IF EXISTS {}',
    }

    create_sql = {
        'knownGene': '''CREATE TABLE {} (
            "name" varchar(255),
            "chrom" varchar(255),
            "strand" char(1),
            "tx_start" integer,
            "tx_end" integer,
            "cds_start" integer,
            "cds_end" integer,
            "exon_count" integer,
            "exon_starts" text,
            "exon_ends" text,
            "protein_id" varchar(40),
            "align_id" varchar(255)
        )''',

        'kgAlias': '''CREATE TABLE {} (
            "kg_id" varchar(40),
            "alias" varchar(80)
        )''',

        'refGene': '''CREATE TABLE {} (
            "bin" integer,
            "name" varchar(255),
            "chrom" varchar(255),
            "strand" char(1),
            "tx_start" integer,
            "tx_end" integer,
            "cds_start" integer,
            "cds_end" integer,
            "exon_count" integer,
            "exon_starts" text,
            "exon_ends" text,
            "score" integer,
            "name2" varchar(255),
            "cds_start_stat" varchar(10),
            "cds_end_stat" varchar(10),
            "exon_frames" text
        )''',

        'ccdsKgMap': '''CREATE TABLE {} (
            "ccds_id" varchar(32) NOT NULL,
            "gene_id" varchar(255) NOT NULL,
            "chrom" varchar(255) NOT NULL,
            "chrom_start" integer NOT NULL,
            "chrom_end" integer NOT NULL,
            "cds_similarity" float NOT NULL
        )''',

        'ccdsExons': '''CREATE TABLE {} (
            "chrom" varchar(255) NOT NULL,
            "start" int NOT NULL,
            "end" int NOT NULL,
            "ccds_id" varchar(255) NOT NULL,
            "exon_id" int NOT NULL,
            "score" int NOT NULL,
            "strand" varchar(1) NOT NULL
        )''',

        'knownToRefSeq': '''CREATE TABLE {} (
            "name" varchar(255) NOT NULL,
            "value" varchar(255) NOT NULL
        )''',
    }

    help = '\n'.join([
        'Performs the various load steps for various UCSC tables including:',
        '- knownGene [1]',
        '- kgAlias [2]',
        '- refGene [3]',
        '- ccdsKgMap [4]',
        '- exons [5]',
        '- knownToRefSeq [6]',
        '',
        'Available targets include: {0}'.format(', '.join(targets)),
        '',
        '[1]: http://genome.ucsc.edu/cgi-bin/hgTables?hgta_doSchemaDb=hg19&hgta_doSchemaTable=knownGene',   # noqa
        '[2]: http://genome.ucsc.edu/cgi-bin/hgTables?hgta_doSchemaDb=hg19&hgta_doSchemaTable=kgAlias',   # noqa
        '[3]: http://genome.ucsc.edu/cgi-bin/hgTables?hgta_doSchemaDb=hg19&hgta_doSchemaTable=refGene',   # noqa
        '[4]: http://genome.ucsc.edu/cgi-bin/hgTables?hgta_doSchemaDb=hg19&hgta_doSchemaTable=ccdsKgMap',   # noqa
        '[5]: File on variome.. /nas/is1/MFalk/framework/ccds_hg19_exons.bed',
        '[6]: http://genome.ucsc.edu/cgi-bin/hgTables?hgta_doSchemaDb=hg19&hgta_doSchemaTable=knownToRefSeq',   # noqa
    ])

    processors = {
        'ccdsExons': CCDSExonsParser,
    }



import vcf
import logging

log = logging.getLogger(__name__)


class StreamEditor(object):
    indices = None

    output_columns = None

    def __init__(self, fin, indices=None, indel='\t', outdel=None,
                 encoding='utf-8'):
        self.fin = fin

        # Input delimiter, output delimiter, default to the same output
        # delimiter if none is defined.
        self.indel = indel
        if outdel is None:
            outdel = indel
        self.outdel = outdel

        # The first iteration will populate the indices if not defined
        self.indices = indices or self.indices

        # Encoding of the input file
        self.encoding = encoding

    def process_column(self, idx, value):
        "Process a single column."
        if value is not None:
            value = str(value).decode(self.encoding)
        return value

    def process_line(self, line):
        "Process a single complete line."
        cleaned = []
        columns = line.split(self.indel)
        # Populate indices if not defined
        if not self.indices:
            self.indices = range(len(columns))
        for i in self.indices:
            # Support turning an in col into multiple out cols
            out = self.process_column(i, columns[i])
            if isinstance(out, (list, tuple)):
                cleaned.extend(out)
            else:
                cleaned.append(out)
        return self.outdel.join(cleaned) + '\n'

    def read(self, size=-1):
        "Reads up to size bytes, but always completes the last line."
        buf = self.fin.read(size)
        if not buf:
            return ''
        lines = buf.splitlines()
        # Read the rest of the last line if necessary
        if not buf.endswith('\n'):
            last = lines.pop()
            partial = self.fin.readline()
            lines.append(last + partial)
        # Process the lines, concatenate them
        lines = [self.process_line(line.rstrip('\n')) for line in lines]
        return ''.join(lines)

    def readline(self, size=-1):
        "The size is ignored since a complete line must be read."
        line = self.fin.readline()
        if not line:
            return ''
        return self.process_line(line.rstrip('\n'))


NULL_VALUES = ('.', '', 'null', 'NA', 'N/A', 'NONE')


class VCFStreamEditor(StreamEditor):
    vcf_fields = []

    info_fields = []

    def __init__(self, *args, **kwargs):
        super(VCFStreamEditor, self).__init__(*args, **kwargs)
        self.reader = vcf.VCFReader(self.fin)

    def _clean_value(self, value):
        if isinstance(value, (tuple, list)):
            return ','.join([str(x) for x in value if x is not None])
        if value is not None:
            value = str(value)
            if value in NULL_VALUES:
                value = None
        return value

    def process_column(self, key, value):
        value = self._clean_value(value)
        return super(VCFStreamEditor, self).process_column(key, value)

    def process_line(self, record):
        "Process a single record. This assumes only a single sample output."
        cleaned = []

        for key in self.vcf_fields:
            out = self.process_column(key, getattr(record, key))
            if isinstance(out, (list, tuple)):
                cleaned.extend(out)
            else:
                cleaned.append(out)

        for key in self.info_fields:
            out = self.process_column(key, record.INFO.get(key, None))
            if isinstance(out, (list, tuple)):
                cleaned.extend(out)
            else:
                cleaned.append(out)

        return cleaned

    def read(self, size=-1):
        """Read `size` bytes from the reader relative to the parsed output.
        This is generally acceptable in practice since VCF lines are
        condensed, but if the output line <<< record, this means the actual
        memory used will be much greater than `size`.
        """
        lines = []
        parsed_size = 0

        while True:
            line = self.readline()
            if not line:
                break
            lines.append(line)
            parsed_size += len(line)
            if size and size > 0 and parsed_size >= size:
                break

        return ''.join(lines)

    def readline(self, size=-1):
        "Ignore the `size` since a complete line must be processed."
        try:
            record = next(self.reader)
            return self.outdel.join(self.process_line(record)) + '\n'
        except StopIteration:
            return ''


class PGCopyEditor(StreamEditor):
    """Ensures NoneTypes are represented as the standard \N string and tabs
    are used as the default delimiter for the output.
    """
    def __init__(self, *args, **kwargs):
        super(PGCopyEditor, self).__init__(*args, **kwargs)
        self.null = kwargs.get('null', '\\N')

    def process_column(self, idx, value):
        value = super(PGCopyEditor, self).process_column(idx, value)
        if value is None:
            value = self.null
        return value


class VCFPGCopyEditor(VCFStreamEditor):
    def __init__(self, *args, **kwargs):
        super(VCFPGCopyEditor, self).__init__(*args, **kwargs)
        self.null = kwargs.get('null', '\\N')

    def process_column(self, key, value):
        value = super(VCFPGCopyEditor, self).process_column(key, value)
        if value is None:
            value = self.null
        return value



import hashlib


def file_md5(f, size=8192):
    "Calculates the MD5 of a file."
    md5 = hashlib.md5()
    while True:
        data = f.read(size)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()



import logging
from django.conf import settings
from django.contrib.auth.models import Group
from django.core.mail import send_mail
from django.db import transaction
from django.db.models.signals import post_save, post_delete
from guardian.shortcuts import assign
from .models import Cohort, Project, Batch, Sample

PROJECT_GROUP_TEMPLATE = '{0} Project Team'
AUTO_PUBLISH_BATCH = getattr(settings, 'VARIFY_AUTO_PUBLISH_BATCH', True)

log = logging.getLogger(__name__)


@transaction.commit_on_success
def update_sample_for_autocreated_cohorts(instance, created, **kwargs):
    "Manages adding/removing samples from autocreated cohorts."

    # World
    lookup = {'batch': None, 'project': None, 'autocreated': True,
              'name': 'World'}
    try:
        world_cohort = Cohort.objects.get(**lookup)
    except Cohort.DoesNotExist:
        if getattr(settings, 'VARIFY_AUTO_CREATE_COHORT', True):
            world_cohort = Cohort(**lookup)
            world_cohort.save()
        else:
            log.info("World cohort was not found and was not created because "
                     "VARIFY_AUTO_CREATE_COHORT setting is False.")
            return

    project = instance.project

    # Project
    lookup = {'batch': None, 'project': project, 'autocreated': True}
    try:
        project_cohort = Cohort.objects.get(**lookup)
    except Cohort.DoesNotExist:
        project_cohort = Cohort(name=unicode(project), **lookup)
        project_cohort.save()

    if instance.published:
        world_cohort.add(instance, added=False)
        project_cohort.add(instance, added=False)
    else:
        world_cohort.remove(instance, delete=True)
        project_cohort.remove(instance, delete=True)


@transaction.commit_on_success
def auto_delete_cohort(instance, **kwargs):
    "Deletes and auto-created cohort named after the instance."
    cohorts = Cohort.objects.filter(autocreated=True)

    if isinstance(instance, Project):
        cohorts = cohorts.filter(project=instance)
    elif isinstance(instance, Batch):
        cohorts = cohorts.filter(batch=instance)
    else:
        return

    count = cohorts.count()
    cohorts.delete()
    log.info('Delete {0} autocreated cohorts for {1}'.format(count, instance))


def auto_create_project_group(instance, created, **kwargs):
    name = PROJECT_GROUP_TEMPLATE.format(instance.name)
    group, created = Group.objects.get_or_create(name=name)

    # TODO change this to queue up an email rather than doing it in process..
    if created:
        assign('view_project', group, instance)

        try:
            kwargs = {
                'subject': '{0}Project "{1}" Created'.format(
                    settings.EMAIL_SUBJECT_PREFIX, name),
                'message': 'The "{0}" Project Group has been created. This is '
                           'a reminder to setup any permissions for the '
                           'associated users.'.format(name),
                'from_email': settings.NO_REPLY_EMAIL,
                'recipient_list': [settings.SUPPORT_EMAIL],
            }
            # Since we aren't passing the fail_silently kwarg to the send_mail
            # method, it will throw any errors back in our face so we catch
            # them here and log them rather than letting it propogate.
            send_mail(**kwargs)
        except Exception:
            log.exception("Error sending project group notification email")

        log.info('Autocreate project group {0}'.format(group))


def auto_delete_project_group(instance, **kwargs):
    name = PROJECT_GROUP_TEMPLATE.format(instance.name)
    Group.objects.filter(name=name).delete()
    log.info('Delete autocreated project group {0}'.format(name))


def update_batch_count(instance, **kwargs):
    """Sample post-save handler to update the sample's batch count.

    Batches are unpublished by default (to prevent publishing empty batches).
    If the `AUTO_PUBLISH_BATCH` setting is true, the batch will be published
    automatically when at least one published sample is in the batch.
    """
    batch = instance.batch
    count = batch.samples.filter(published=True).count()

    if count != batch.count:
        batch.count = count
        if AUTO_PUBLISH_BATCH:
            batch.published = bool(count)
        batch.save()


post_save.connect(update_sample_for_autocreated_cohorts, sender=Sample)
post_save.connect(update_batch_count, sender=Sample)
post_save.connect(auto_create_project_group, sender=Project)

post_delete.connect(auto_delete_cohort, sender=Batch)
post_delete.connect(auto_delete_cohort, sender=Project)



from vdw.raw.management.subcommands.load import LoadCommand


class Command(LoadCommand):

    targets = ['genePhenotypes']

    qualified_names = {
        'genePhenotypes': '"raw"."hpo_gene_phenotypes"',
    }

    drop_sql = {
        'genePhenotypes': 'DROP TABLE IF EXISTS {}',
    }

    create_sql = {
        'genePhenotypes': '''CREATE TABLE {} (
            "entrez_gene_id" integer NOT NULL,
            "entrez_gene_symbol" varchar(20) NOT NULL,
            "hpo_terms" text NOT NULL
        )''',
    }



from django.db import models, IntegrityError


class GeneManager(models.Manager):
    def find(self, symbol, chrom=None, create=False):
        """
        Find a gene based on the symbol or disambiguate using synonyms.
        If no gene is found, if the create is True, a new instance will be
        created with that symbol.
        """
        queryset = self.get_query_set()

        # Filter by chromosome if one is specified
        if chrom:
            queryset = queryset.filter(chr=chrom)

        # Look for direct match
        matches = queryset.filter(symbol__iexact=symbol)

        try:
            return matches[0]
        except IndexError:
            pass

        # Attempt to disambiguate, only if this is the only synonym may it be
        # associated
        potential = list(
            queryset.filter(synonyms__label__iexact=symbol).distinct())

        if len(potential) == 1:
            return potential[0]

        # Only if there are no matches should we create a new record,
        # otherwise the synonym war will continue
        if create and len(potential) == 0:
            instance = self.model(chr=chrom, symbol=symbol)
            try:
                instance.save()
            except IntegrityError:
                instance = queryset.get(chr=chrom, symbol=symbol)
            return instance



import logging
import os
import sys
import shlex
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from ConfigParser import ConfigParser
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.db import connections, transaction, DEFAULT_DB_ALIAS
from vdw.raw.utils.stream import PGCopyEditor

log = logging.getLogger(__name__)


class LoadCommand(BaseCommand):

    requires_model_validation = False

    # a list of 'targets' which act as unique identifiers to tables and
    # manifest entries
    targets = []

    # schema.table
    qualified_names = {}

    # CREATE TABLE {} ( ... )
    create_sql = {}

    default_processor = PGCopyEditor

    processors = {}

    option_list = BaseCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Nominates a database to print the SQL for. Defaults '
                         'to the "default" database.'),
        make_option('-s', '--source', action='store', dest='source',
                    default='.', help='Specifies the directory containing the '
                                      'MANIFEST and raw data files.'),
        make_option('-d', '--drop', action='store_true', default=False,
                    help='Only perform the DROP TABLE operation'),
        make_option('-c', '--create', action='store_true', default=False,
                    help='Only perform the CREATE TABLE operation'),
        make_option('-l', '--load', action='store_true', default=False,
                    help='Only perform the data loading (into the existing '
                         'tables)')
    )

    def drop_tables(self, cursor, targets):
        log.debug('Dropping tables...')
        for target in targets:
            try:
                self.drop_table(cursor, target, targets[target][1])
                log.debug('- {0}'.format(target))
            except Exception:
                log.exception(
                    'Error dropping {0}, rolling back..'.format(target))
                transaction.rollback()
                sys.exit(1)

    def create_tables(self, cursor, targets):
        log.debug('Creating tables...')
        for target in targets:
            try:
                self.create_table(cursor, target, targets[target][1])
                log.debug('- {0}'.format(target))
            except Exception:
                log.exception(
                    'Error creating {0}, rolling back'.format(target))
                transaction.rollbak()
                sys.exit(1)

    def load_tables(self, cursor, targets):
        log.debug('Loading tables...')
        for target in targets:
            files, options = targets[target]
            try:
                self.load_files(cursor, target, files, options)
                log.debug('- {0}'.format(target))
            except Exception:
                log.exception(
                    'Error loading {0}, rolling back'.format(target))
                transaction.rollback()
                sys.exit(1)

    def drop_table(self, cursor, target, options):
        "Drops the target table."
        sql = 'DROP TABLE IF EXISTS {0}'
        cursor.execute(sql.format(self.qualified_names[target]))

    def create_table(self, cursor, target, options):
        "Creates the target table."
        cursor.execute(
            self.create_sql[target].format(self.qualified_names[target]))

    def load_files(self, cursor, target, files, options):
        "Loads multiple files into the target table."
        for fname in files:
            self.load_file(cursor, target, fname, options)

    def load_file(self, cursor, target, fname, options):
        "Parses and loads a single file into the target table."
        with open(fname) as fin:
            log.debug("opening {0} in {1} load_file".format(fname, __name__))
            encoding = options.get('encoding', 'utf-8')
            if target in self.processors:
                reader = self.processors[target](fin, encoding=encoding)
            else:
                reader = self.default_processor(fin, encoding=encoding)
            columns = getattr(reader, 'output_columns', None)
            for _ in xrange(int(options.get('skip-lines', 0))):
                fin.readline()
            cursor.copy_from(reader, self.qualified_names[target],
                             columns=columns)

    def handle(self, **options):
        if not self.targets:
            log.info('No targets defined. Nothing to do.')
            return

        source = options.get('source')
        drop = options.get('drop')
        create = options.get('create')
        load = options.get('load')
        using = options.get('database')
        connection = connections[using]
        cursor = connection.cursor()

        # If no operation is supplied, assume all operations
        if not (drop or create or load):
            drop = create = load = True

        # Perform checks ahead of time to prevent failing in an
        # inconsistent state
        if not os.path.exists(os.path.join(source, 'MANIFEST')):
            raise CommandError('No MANIFEST in the source directory: {0}'
                               .format(source))

        targets = OrderedDict()
        parser = ConfigParser()
        parser.read([os.path.join(source, 'MANIFEST')])

        for target in parser.sections():
            if target not in self.targets:
                log.warn('Unknown target "{0}", skipping...'.format(target))
                continue

            options = dict(parser.items(target))

            files = []
            for fn in shlex.split(options['files']):
                path = os.path.join(source, fn)
                if not os.path.exists(path):
                    raise CommandError('No file named {0} exists'.format(fn))
                files.append(path)
            targets[target] = (files, options)

        if not targets:
            log.info('No targets defined in MANIFEST. Nothing to do.')
            return

        # Manually commit along the way
        with transaction.commit_manually(using=using):
            if drop:
                self.drop_tables(cursor, targets)

            if create:
                self.create_tables(cursor, targets)

            if load:
                self.load_tables(cursor, targets)

            transaction.commit()



import logging
import tempfile
from cStringIO import StringIO
from django.db import transaction

log = logging.getLogger(__name__)

DEFAULT_BATCH_SIZE = 1000


def writetmp(stream):
    tmp = tempfile.NamedTemporaryFile()
    while True:
        buff = stream.read(2 ** 16)
        if not buff:
            break
        tmp.write(buff)

    # Flush to disk.. seek to top for COPY
    tmp.flush()
    tmp.seek(0)
    return tmp


def pgcopy(stream, db_table, columns, cursor, database):
    with transaction.commit_on_success(database):
        try:
            cursor.copy_from(stream, db_table, columns=columns)
        except Exception as e:
            log.exception(e)
            raise


def batch_stream(buff, stream, size=DEFAULT_BATCH_SIZE):
    """Writes a batch of `size` lines to `buff`.

    Returns boolean of whether the stream has been exhausted.
    """
    buff.truncate(0)
    for _ in xrange(size):
        if hasattr(stream, 'readline'):
            line = stream.readline()
        else:
            try:
                line = next(stream)
            except StopIteration:
                line = ''

        # No more lines, return the tmp
        if line == '':
            buff.seek(0)
            return True
        buff.write(line)
    buff.seek(0)
    return False


def pgcopy_batch(stream, db_table, columns, cursor, database,
                 batch_size=DEFAULT_BATCH_SIZE):
    buff = StringIO()
    try:
        while True:
            empty = batch_stream(buff, stream, batch_size)
            pgcopy(buff, db_table, columns, cursor, database)
            if empty:
                break
    finally:
        buff.close()



import logging
import re
from optparse import make_option
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from django.db import connections, DEFAULT_DB_ALIAS
from django.core.management.base import BaseCommand
from vdw.genes.models import GeneFamily, Gene, Synonym
from vdw.genome.models import Chromosome
from vdw.literature.models import PubMed

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Cleans and loads genes from various sources into a uniform set of tables.
    """
    option_list = BaseCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Nominates a database to print the SQL for. Defaults '
                         'to the "default" database.'),
        make_option('--genes', action='store_true', default=False,
                    help='Reload genes, synonyms and PubMed IDs from HGNC'),
        make_option('--families', action='store_true', default=False,
                    help='Reload gene families provided by HGNC'),
    )

    def _get_or_create_gene(self, record):
        target = {}

        # Link HGNC id
        target['hgnc_id'] = int(record['hgnc_id'])

        # Parse and map chromosome
        if record['chromosome'] == 'mitochondria':
            target['chr_id'] = self.chromosomes['MT']
        elif ' and ' in record['chromosome']:
            target['chr_id'] = self.chromosomes['XY']
        else:
            match = self.chrom_re.match(record['chromosome'])
            if not match:
                log.warning('unable to match gene chromosome from HGNC',
                            extra={
                                'hgnc_id': record['hgnc_id'],
                                'raw_chr': record['chromosome'],
                            })
                return None, False
            target['chr_id'] = self.chromosomes[match.groups()[0]]

        target['symbol'] = record['approved_symbol'].encode('utf8')
        target['name'] = record['approved_name'].encode('utf8')

        # If the gene already exists by hgnc_id, fetch it. Next check
        # by symbol (in case of a new approved gene) and set the
        # hgnc_id. Fallback to creating a new gene
        try:
            return Gene.objects.get(hgnc_id=target['hgnc_id']), False
        except Gene.DoesNotExist:
            try:
                gene = Gene.objects.get(symbol=target['symbol'])
                gene.hgnc_id = target['hgnc_id']
            except Gene.DoesNotExist:
                gene = Gene(**target)
            gene.save()
            return gene, True

    def _get_or_create_synonym(self, synonym):
        try:
            return Synonym.objects.get(label=synonym), False
        except Synonym.DoesNotExist:
            synonym = Synonym(label=synonym)
            synonym.save()
            return synonym, True

    def _get_or_create_article(self, pmid):
        try:
            return PubMed.objects.get(pmid=pmid), False
        except PubMed.DoesNotExist:
            article = PubMed(pmid=pmid)
            article.save()
            return article, True

    def load_genes(self, cursor):
        # The columns we care about
        keys = ['hgnc_id', 'approved_symbol', 'approved_name',
                'previous_symbols', 'previous_names', 'synonyms',
                'name_synonyms', 'chromosome', 'pubmed_ids']

        cursor.execute('''
            SELECT DISTINCT "hgnc_id", "approved_symbol", "approved_name",
                            "previous_symbols", "previous_names", "synonyms",
                            "name_synonyms", "chromosome", "pubmed_ids"
            FROM "raw"."hgnc"
            WHERE "status" = 'Approved'
                AND "chromosome" != 'reserved'
            ORDER BY "approved_symbol"
        ''')

        self.chromosomes = dict(Chromosome.objects.values_list('value', 'id'))
        self.chrom_re = re.compile(r'^c?([\dXY]{1,2})')

        count = 0
        new_genes = 0
        new_synonyms = 0
        new_articles = 0

        while True:
            rows = cursor.fetchmany(100)
            if not rows:
                break

            for row in rows:
                count += 1
                source_record = OrderedDict(zip(keys, row))

                # Get the gene
                gene, created = self._get_or_create_gene(source_record)

                if not gene:
                    continue

                if created:
                    new_genes += 1

                # Parse alternative symbols and names
                previous_symbols = source_record['previous_symbols'] or ''
                previous_names = source_record['previous_names'] or ''
                synonyms = source_record['synonyms'] or ''
                name_synonyms = source_record['name_synonyms'] or ''

                # Add approved as aliases for ease of querying a single source
                approved = [gene.symbol, gene.name]

                synonyms = ', '.join(
                    [previous_symbols, previous_names, synonyms,
                     name_synonyms]).split(', ') + approved
                for synonym in synonyms:
                    synonym = synonym.strip('" ').encode('utf8')
                    synonym, created = self._get_or_create_synonym(synonym)
                    if created:
                        new_synonyms += 1
                    gene.synonyms.add(synonym)

                # Associate PubMed IDs
                pubmed_ids = source_record['pubmed_ids'] or ''

                for pmid in set(pubmed_ids.split(', ')):
                    try:
                        pmid = int(pmid)
                    except (ValueError, TypeError):
                        continue
                    article, created = self._get_or_create_article(pmid)
                    if created:
                        new_articles += 1
                    gene.articles.add(article)

                log.debug(
                    'scanned: {0}\tgenes: {1}\tsynonyms: {2}\tarticles: {3}'
                    .format(count, new_genes, new_synonyms, new_articles))

    def load_families(self, cursor):
        cursor.execute('''
            SELECT tag, description, gene.id
            FROM raw.hgnc_families
                LEFT OUTER JOIN gene ON (hgnc_families.hgnc_id = gene.hgnc_id)
            ORDER BY tag
        ''')

        keys = ['tag', 'description', 'gene_id']

        families = dict((f.tag, f) for f in GeneFamily.objects.all())

        while True:
            rows = cursor.fetchmany(100)
            if not rows:
                break

            for row in rows:
                record = dict(zip(keys, row))

                if record['tag'] not in families:
                    family = GeneFamily(tag=record['tag'],
                                        description=record['description'])
                    family.save()
                    families[family.tag] = family

                if record['gene_id']:
                    gene = Gene(pk=record['gene_id'])
                    family = families[record['tag']]
                    gene._state.db = family._state.db
                    gene.families.add(family)

    def handle(self, *args, **options):
        using = options.get('database')
        load_genes = options.get('genes')
        load_families = options.get('families')

        connection = connections[using]
        cursor = connection.cursor()

        if not any([load_genes, load_families]):
            log.debug('Nothing to do.')
            return

        if load_genes:
            self.load_genes(cursor)
        if load_families:
            self.load_families(cursor)



import os
import logging
import sts
import time
import vcf
from django.conf import settings
from django.db import connections, transaction
from sts.contextmanagers import transition
from vdw.pipeline import checks, job, ManifestReader
from vdw.pipeline.load import pgcopy_batch
from vdw.samples.models import Result, Project, Batch, Sample, \
    SampleManifest, DEFAULT_PROJECT_NAME
from . import SAMPLE_CHANNEL
from .utils import ResultStream

log = logging.getLogger(__name__)


@transaction.commit_on_success
def create_sample(sample_name, vcf_colname, batch_name, project_name=None,
                  version=0):
    version = int(version)

    if project_name:
        # Ensure a project is available for the sample
        project, created = Project.objects.get_or_create(
            name__iexact=project_name,
            defaults={'label': project_name, 'name': project_name})
    else:
        project = Project.objects.get(name=DEFAULT_PROJECT_NAME)

    batch, created = Batch.objects.get_or_create(
        name__iexact=batch_name, project=project,
        defaults={'label': batch_name, 'name': batch_name})

    sample, created = Sample.objects.get_or_create(
        name__iexact=sample_name, project=project, batch=batch,
        version=version, defaults={'label': sample_name, 'name': sample_name,
                                   'vcf_colname': vcf_colname})

    return sample, created


def check_sample_section(manifest):
    section = manifest.section('sample')
    # Sample itself is no longer a required key
    required_keys = ('batch', 'project', 'version')
    return all([key in section for key in required_keys])


@job(SAMPLE_CHANNEL)
def load_samples(manifest_path, database, **kwargs):
    manifest = ManifestReader(manifest_path)

    genome_version = settings.VDW_GENOME_VERSION
    if genome_version:
        genome_info = manifest.section('genome')

        if not genome_info:
            log.warn('Genome version "{0}" is required but genome section is '
                     'not included in MANIFEST({1}). Skipping sample.'
                     .format(genome_version, manifest_path))
            return

        if genome_version != genome_info['version']:
            log.warn(
                'Genome version "{0}" is required but version "{1}" was found '
                'in the MANIFEST({2}). Skipping sample.'
                .format(genome_version, genome_info['version'], manifest_path))
            return

    # Ensure the sample is marked to be loaded.
    if not manifest.marked_for_load():
        log.info('Sample not marked for load', extra={
            'manifest_path': manifest_path,
        })
        return

    # Ensure the sample section is valid.
    if not check_sample_section(manifest):
        log.warn('Manifest sample section is not valid', extra={
            'manifest_path': manifest_path,
        })
        return

    sample_info = manifest.section('sample')
    vcf_info = manifest.section('vcf')

    # ignore whatever sample is listed in the manifest and scan the vcf
    # for samples
    vcf_path = os.path.join(os.path.dirname(manifest_path), vcf_info['file'])

    with open(vcf_path) as file_obj:
        log.debug('opening {0} in load_samples'.format(vcf_path))
        reader = vcf.Reader(file_obj)
        samples = reader.samples

    if 'sample' in sample_info:
        pretty_names = sample_info['sample'].split(',')
    else:
        pretty_names = samples

    if len(samples) != len(pretty_names):
        log.warn('Length of comma-delimited samples field in manifest '
                 'does not match the length of samples in {0}'
                 .format(vcf_info['file']))
        return

    # Create the sample (and batch and project if needed)..
    num_created = 0
    num_skipped = 0

    for pretty_name, vcf_sample in zip(pretty_names, samples):
        log.debug('Trying to create {0} sample record'.format(vcf_sample))
        sample, created = create_sample(sample_name=pretty_name,
                                        vcf_colname=vcf_sample,
                                        batch_name=sample_info['batch'],
                                        project_name=sample_info['project'],
                                        version=sample_info['version'])
        log.debug('{0} created'.format(sample))

        if created:
            num_created += 1
            sts.transition(sample, 'Sample Record Created')
        else:
            num_skipped += 1

        manifest = SampleManifest.objects.filter(sample=sample)
        # Create a manfiest object for the sample if one does not exist
        if created or not manifest.exists():
            sample_manifest = SampleManifest(sample=sample)
            sample_manifest.load_content(manifest_path)
            sample_manifest.save()
            sts.transition(sample, 'Sample Manifest Created')

    # Publish to channel that this manifest is eligible for processing
    # downstream..
    if num_created > 0 or kwargs.get('force', False):
        SAMPLE_CHANNEL.publish(manifest_path=manifest_path, database=database)

    # Returns whether the sample has been created
    load_dict = {'created': num_created, 'skipped': num_skipped}
    return load_dict


@job(timeout=60 * 60)
def load_results(manifest_path, database, **kwargs):
    manifest = ManifestReader(manifest_path)

    if not manifest.marked_for_load():
        log.info('Sample not marked for load', extra={
            'manifest_path': manifest_path,
        })
        return

    # Ensure the sample section is valid..
    if not check_sample_section(manifest):
        log.warn('Manifest sample section is not valid', extra={
            'manifest_path': manifest_path,
        })
        return

    sample_info = manifest.section('sample')
    vcf_info = manifest.section('vcf')
    # Ignore whatever sample is listed in the manifest and scan the vcf for
    # samples.
    vcf_path = os.path.join(os.path.dirname(manifest_path), vcf_info['file'])
    with open(vcf_path) as file_obj:
        log.debug('opening {0} in load_samples'.format(vcf_path))
        reader = vcf.Reader(file_obj)
        samples = reader.samples
    if 'sample' in sample_info:
        pretty_names = sample_info['sample'].split(',')
    else:
        pretty_names = samples

    for pretty_name, vcf_sample in zip(pretty_names, samples):
        try:
            sample = Sample.objects.get(
                name__iexact=pretty_name,
                batch__name__iexact=sample_info['batch'],
                project__name__iexact=sample_info['project'],
                version=sample_info['version'])
        except Sample.DoesNotExist:
            log.error('Sample does not exist', extra=sample_info)
            return

        # Is it already loaded, let's skip for now
        if Result.objects.filter(sample=sample).exists():
            log.debug('{0} exists in results'.format(vcf_sample))
        else:
            log.debug('about to load results for {0}'.format(vcf_sample))

            # STSError: Cannot start transition while already in one.
            successful = False
            while not successful:
                try:
                    with transition(sample, 'Sample Published',
                                    event='Loading Results'):
                        connection = connections[database]
                        cursor = connection.cursor()

                        with open(vcf_path) as fin:
                            stream = ResultStream(fin, sample_id=sample.id,
                                                  vcf_sample=vcf_sample)
                            columns = stream.output_columns
                            db_table = Result._meta.db_table
                            pgcopy_batch(stream, db_table, columns, cursor,
                                         database)

                        # Update result count
                        sample.count = sample.results.count()
                        sample.published = True
                        sample.save()
                        successful = True
                except:
                    log.exception('STS errors')
                    time.sleep(10)

    vcf_info = manifest.section('vcf')

    # Absolute path relative to the MANIFEST directory
    vcf_path = os.path.join(os.path.dirname(manifest_path), vcf_info['file'])

    # Compare expected MD5 (in manifest) to the file MD5
    if 'md5' in vcf_info:
        vcf_md5 = checks.file_md5(vcf_path)

        if vcf_md5 != vcf_info['md5']:
            log.error('VCF file MD5 does not match expected in manifest',
                      extra={'manifest_path': manifest_path})

    # Existing samples by the same name of a previous version are unpublished
    # now that is is ready to be published.
    count = Sample.objects.filter(
        name__iexact=sample.name, project=sample.project, batch=sample.batch,
        version__lt=sample.version).update(published=False)

    if count:
        log.info('{0} previous versions unpublished for {1}'
                 .format(count, sample))



DEFAULT_PRIORITY_SCORE = 0.7
MIN_COVERAGE_DEPTH = 10

PRIORITY_FEATURE_WEIGHTS = {
    'non_candidate_gene': 0.4,
    'unknown_variant': 0.3,
    'no_hgmd_entry': 0.3,
    'low_coverage_depth': 0.3,
    'synonymous': 0.1,
    'intronic': 0,
}

PRIORITY_FEATURES = sorted(PRIORITY_FEATURE_WEIGHTS.keys())


def calculate_priority(ratios=None, **kwargs):
    "Calculates a priority score based on a number of attributes."
    if not ratios:
        ratios = PRIORITY_FEATURE_WEIGHTS
    scores = [DEFAULT_PRIORITY_SCORE]
    for key, value in kwargs.items():
        if key not in PRIORITY_FEATURE_WEIGHTS:
            raise KeyError('The following keyword arguments are supported: '
                           '{keys}'.format(keys=PRIORITY_FEATURES))
        if value is True:
            scores.append(PRIORITY_FEATURE_WEIGHTS[key])
    return float(sum(scores)) / len(scores)



import logging
from optparse import make_option
from django.db import models, transaction
from django.db.models import loading
from django.conf import settings
from django.db import connections, DEFAULT_DB_ALIAS
from django.db.utils import DatabaseError
from django.core.management.base import NoArgsCommand
from django.template.defaultfilters import filesizeformat

log = logging.getLogger(__name__)


class Command(NoArgsCommand):
    help = 'Prints the calculated data size of each table on disk.'

    option_list = NoArgsCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Nominates a database to print the '
                         'SQL for.  Defaults to the "default" database.'),
        make_option('-s', action='store_true', dest='suffix',
                    help='Prints human-readable sizes with the approiate '
                         'suffix.'),
    )

    def handle_noargs(self, **options):
        using = options.get('database')
        suffix = options.get('suffix')
        db_name = settings.DATABASES[using]['NAME']
        cursor = connections[using].cursor()
        loading.cache._populate()

        sizes = []
        with transaction.commit_on_success(using):
            for name, app in loading.cache.app_store.items():
                for model in models.get_models(app):
                    opts = model._meta
                    # Some tables may not yet be created in the database, so
                    # this statement will fail
                    try:
                        cursor.execute("SELECT pg_total_relation_size('{0}')"
                                       .format(opts.db_table))
                        sizes.append((cursor.fetchone()[0], opts.module_name))
                    except DatabaseError:
                        transaction.rollback()
        for size, name in sorted(sizes):
            if suffix:
                size = filesizeformat(size)
            log.debug('{0}\t{1}'.format(size, name))
        if suffix:
            cursor.execute("SELECT pg_size_pretty(pg_database_size('{0}'))"
                           .format(db_name))
        else:
            cursor.execute("SELECT pg_database_size('{0}')".format(db_name))
        log.debug('{0}\tTOTAL'.format(cursor.fetchone()[0]))



from django.db import models
from avocado.models import Lexicon
from vdw.genome.models import Chromosome
from vdw.literature.models import PubMed
from vdw.phenotypes.models import Phenotype, PhenotypeThrough
from vdw.genes.models import Gene, Exon, Transcript
from .managers import VariantManager
from .utils import calculate_md5


class EffectImpact(Lexicon):
    "Impact of an effect, e.g. high, moderate, low, etc."
    label = models.CharField(max_length=100)
    value = models.CharField(max_length=100)

    class Meta(Lexicon.Meta):
        db_table = 'effect_impact'


class EffectRegion(Lexicon):
    "Region of the genome an effect can affect."
    label = models.CharField(max_length=100)
    value = models.CharField(max_length=100)

    class Meta(Lexicon.Meta):
        db_table = 'effect_region'


class Effect(Lexicon):
    "The effect a variant can have."
    label = models.CharField(max_length=100)
    value = models.CharField(max_length=100)
    # http://snpeff.sourceforge.net/faq.html#What_effects_are_predicted?
    description = models.TextField(null=True, blank=True)
    impact = models.ForeignKey(EffectImpact, null=True, blank=True)
    region = models.ForeignKey(EffectRegion, null=True, blank=True)

    class Meta(Lexicon.Meta):
        db_table = 'effect'


class VariantType(Lexicon):
    "Lexicon of genetic variant types."
    label = models.CharField(max_length=20)
    value = models.CharField(max_length=20)

    class Meta(Lexicon.Meta):
        db_table = 'variant_type'


class Variant(models.Model):
    chr = models.ForeignKey(Chromosome)
    pos = models.IntegerField('position')
    ref = models.TextField('reference', db_index=True)
    alt = models.TextField('alternate', db_index=True)
    md5 = models.CharField(max_length=32, editable=False)

    rsid = models.TextField('dbSNP ID', null=True, blank=True)

    # Note, when loading the data via the VCF file, this is value is set for
    # the entire file, e.g. indel or SNP
    type = models.ForeignKey(VariantType, null=True)

    # Track whether this variant has gone through liftover
    liftover = models.NullBooleanField(
        'has been lifted over?', blank=True, editable=False)

    # Literature
    articles = models.ManyToManyField(PubMed, db_table='variant_pubmed')

    # Phenotypes
    phenotypes = models.ManyToManyField(Phenotype, through='VariantPhenotype')

    objects = VariantManager()

    class Meta(object):
        db_table = 'variant'
        unique_together = ('chr', 'pos', 'ref', 'alt')

    def natural_key(self):
        return (self.chr, self.pos, self.ref, self.alt)

    @classmethod
    def calculate_md5(cls, chr, pos, ref, alt):
        return calculate_md5(chr, pos, ref, alt)

    def ncbi_url(self):
        if self.rsid:
            return 'http://www.ncbi.nlm.nih.gov/projects/SNP/snp_ref.cgi?rs={}' + self.rsid[2:]     # noqa


class VariantPhenotype(PhenotypeThrough):
    variant = models.ForeignKey(Variant, related_name='variant_phenotypes')

    class Meta(object):
        db_table = 'variant_phenotype'


class ThousandG(models.Model):
    variant = models.ForeignKey(Variant, related_name='thousandg', unique=True)
    an = models.IntegerField('total allele count', null=True)
    ac = models.IntegerField('alternate allele count', null=True)
    af = models.FloatField('global allele frequency based on AC/AN',
                           null=True, db_index=True)
    aa = models.TextField('ancestral allele', null=True)
    amr_af = models.FloatField('allele frequency for AMR (American)',
                               null=True, db_index=True)
    asn_af = models.FloatField('allele frequency for ASN (Asian)',
                               null=True, db_index=True)
    afr_af = models.FloatField('allele frequency for AFR (African)',
                               null=True, db_index=True)
    eur_af = models.FloatField('allele frequency for EUR (European)',
                               null=True, db_index=True)

    class Meta(object):
        db_table = '1000g'


class EVS(models.Model):
    variant = models.ForeignKey(Variant, related_name='evs', unique=True)
    ea_ac_ref = models.CharField('reference allele count for EUR (European)',
                                 max_length=20, null=True)
    ea_ac_alt = models.CharField('alternate allele count for EUR (European)',
                                 max_length=20, null=True)
    aa_ac_ref = models.CharField('reference allele count for AFR (African)',
                                 max_length=20, null=True)
    aa_ac_alt = models.CharField('alternate allele count for AFR (African)',
                                 max_length=20, null=True)
    all_ac_ref = models.CharField('reference allele count combined',
                                  max_length=20, null=True)
    all_ac_alt = models.CharField('alternate allele count combined',
                                  max_length=20, null=True)
    ea_af = models.FloatField('allele frequency (AF) for EUR (European)',
                              null=True, db_index=True)
    aa_af = models.FloatField('allele frequency (AF) for AFR (African)',
                              null=True, db_index=True)
    all_af = models.FloatField('allele frequency (AF) combined', null=True,
                               db_index=True)
    gts = models.CharField(
        'observed genotypes', max_length=200, null=True,
        help_text='For INDELs, A1, A2, or An refers to the N-th alternate '
                  'allele while R refers to the reference allele.')
    ea_gtc = models.CharField('genotype counts for EUR (European)',
                              max_length=200, null=True)
    aa_gtc = models.CharField('genotype counts for AFR (African)',
                              max_length=200, null=True)
    all_gtc = models.CharField('genotype counts combined', max_length=200,
                               null=True)
    read_depth = models.IntegerField('read depth', null=True)
    clinical_association = models.TextField(null=True)

    class Meta(object):
        db_table = 'evs'


class PolyPhen2(models.Model):
    variant = models.ForeignKey(Variant, related_name='polyphen2')
    score = models.FloatField('PolyPhen2 score', null=True, db_index=True)
    refaa = models.CharField('reference amino acid', max_length=2, null=True)

    class Meta(object):
        db_table = 'polyphen2'

    @classmethod
    def get_prediction(cls, score):
        if score is None:
            return
        if score <= 0.2:
            return 'Benign'
        elif 0.2 < score < 0.85:
            return 'Possibly Damaging'
        elif 0.85 <= score:
            return 'Probably Damaging'

    @property
    def prediction(self):
        return self.get_prediction(self.score)


class Sift(models.Model):
    variant = models.ForeignKey(Variant, related_name='sift')
    score = models.FloatField('SIFT score', null=True, db_index=True)
    refaa = models.CharField('reference amino acid', max_length=2, null=True)
    varaa = models.CharField('alternate amino acid', max_length=2, null=True)

    class Meta(object):
        db_table = 'sift'

    @classmethod
    def get_prediction(cls, score):
        if score is None:
            return
        if score <= 0.5:
            return 'Damaging'
        elif score > 0.5:
            return 'Tolerated'

    @property
    def prediction(self):
        return self.get_prediction(self.score)


class FunctionalClass(Lexicon):
    label = models.CharField(max_length=100)
    value = models.CharField(max_length=100)

    class Meta(Lexicon.Meta):
        db_table = 'variant_functional_class'
        verbose_name_plural = 'functional classes'


class VariantEffect(models.Model):
    "A variant may have one or more effects associated."
    variant = models.ForeignKey(Variant, null=True, blank=True,
                                related_name='effects')

    codon_change = models.TextField(null=True, blank=True)
    amino_acid_change = models.TextField(null=True, blank=True)

    # If the variant falls within a particular exon, exon will be populated. If
    # it does not but applies to a particular gene transcript, transcript will
    # be populated. If neither of these are available, both values will be null
    exon = models.ForeignKey(Exon, null=True, blank=True)
    transcript = models.ForeignKey(Transcript, null=True, blank=True)
    gene = models.ForeignKey(Gene, null=True, blank=True)
    effect = models.ForeignKey(Effect, null=True, blank=True)
    functional_class = models.ForeignKey(FunctionalClass, null=True,
                                         blank=True)

    hgvs_c = models.CharField('HGVS Coding DNA', max_length=200, null=True,
                              blank=True, db_index=True)
    hgvs_p = models.CharField('HGVS Protein', max_length=200, null=True,
                              blank=True, db_index=True)
    segment = models.CharField(max_length=200, null=True, blank=True)

    class Meta(object):
        db_table = 'variant_effect'



import hashlib
from vcf.model import _Record

MD5_ARG_DELIMITER = '|'


def calculate_md5(*args):
    if len(args) == 1 and isinstance(args[0], _Record):
        r = args[0]
        values = [r.CHROM, r.POS, r.REF, ','.join([str(x) for x in r.ALT])]
    elif len(args) == 4:
        values = list(args)
    else:
        raise ValueError('Invalid arguments')
    for i, value in enumerate(values):
        # Position is an int, everything else a string
        if i == 1:
            assert isinstance(value, int) or value.isdigit()
            values[i] = str(value)
        else:
            assert isinstance(value, basestring)
    return hashlib.md5(MD5_ARG_DELIMITER.join(values)).hexdigest()



import logging
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.db import connections, transaction, DEFAULT_DB_ALIAS, DatabaseError
from vdw.samples.models import Sample
from vdw.signals import pre_delete_sample

log = logging.getLogger(__name__)


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Specifies the target database to delete the sample '
                         'from.'),
    )

    def handle(self, *args, **options):
        if not args:
            raise CommandError('A sample ID must be specified')

        database = options.get('database')
        cursor = connections[database].cursor()

        invalid_ids = set(Sample.objects.filter(
            results__assessment__isnull=False, pk__in=args).values_list(
            'id', flat=True))

        if invalid_ids:
            log.warning("Sample(s) with id '{0}' cannot be deleted "
                        "because they have knowledge capture data "
                        "associated with them. If you are absolutely "
                        "sure you want to delete these samples, "
                        "manually delete the knowledge capture data "
                        "for them and then rerun this command."
                        .format(", ".join([str(id) for id in invalid_ids])))

        valid_ids = [id for id in args if int(id) not in invalid_ids]
        columns = ", ".join(['%s' for id in valid_ids])

        if not valid_ids:
            return

        with transaction.commit_manually(database):
            try:
                # Remove sample from cohorts
                cursor.execute('''
                    DELETE FROM cohort_sample WHERE sample_id IN ({0})
                '''.format(columns), valid_ids)

                # Remove result scores
                cursor.execute('''
                    DELETE FROM result_score WHERE result_id IN
                        (SELECT id from sample_result WHERE sample_id IN ({0}))
                '''.format(columns), valid_ids)

                # Remove results of sample
                cursor.execute('''
                    DELETE FROM sample_result WHERE sample_id IN ({0})
                '''.format(columns), valid_ids)

                # Remove sample manifest
                cursor.execute('''
                    DELETE FROM sample_manifest WHERE sample_id IN ({0})
                '''.format(columns), valid_ids)

                pre_delete_sample.send(
                    sender=Sample, pk_set=valid_ids, cursor=cursor)

                cursor.execute('''
                    DELETE FROM sample WHERE sample.id IN ({0})
                '''.format(columns), valid_ids)

                transaction.commit()
            except DatabaseError:
                transaction.rollback()
                log.exception("Error occurred while deleting sample(s) with "
                              "id '{0}'".format(valid_ids))
                raise



import logging
from optparse import make_option
from django.db import transaction, connections, DEFAULT_DB_ALIAS
from django.core.management.base import BaseCommand
from vdw.variants.models import Transcript, VariantEffect
from vdw.variants.pipeline.utils import EffectStream
from vdw.pipeline.load import pgcopy_batch

log = logging.getLogger(__name__)


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Nominates a database to print the SQL for. Defaults '
                         'to the "default" database.'),

        make_option('--transcripts', action='store_true', default=False,
                    help='Causes the transcript table to be truncated prior '
                         'to reloading.'),

        make_option('--stdout', action='store_true', default=False,
                    help='Writes the stream to stdout rather than to the '
                         'database'),
    )

    def handle(self, path, **options):
        database = options.get('database')
        transripts = options.get('transripts')
        stdout = options.get('stdout')

        with open(path) as fin:
            stream = EffectStream(fin, skip_existing=False)

            if stdout:
                while True:
                    line = stream.readline()
                    if line == '':
                        break
                    log.debug(line)
            else:
                cursor = connections[database].cursor()

                with transaction.commit_manually(database):
                    try:
                        cursor.execute('TRUNCATE {0}'.format(
                            VariantEffect._meta.db_table))
                        if transripts:
                            cursor.execute('TRUNCATE {0} CASCADE'.format(
                                Transcript._meta.db_table))
                        columns = stream.output_columns
                        db_table = VariantEffect._meta.db_table
                        pgcopy_batch(stream, db_table, columns, cursor,
                                     database)

                        transaction.commit(database)
                    except Exception as e:
                        transaction.rollback(database)
                        log.exception(e)
                        raise



import os
import logging
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from django.db import connections, transaction
from vdw import db
from vdw.variants.models import Variant, VariantEffect, EVS, ThousandG, \
    Sift, PolyPhen2
from vdw.pipeline import job, ManifestReader
from vdw.pipeline.load import pgcopy_batch
from .utils import VariantStream, EffectStream
from . import VARIANT_CHANNEL

log = logging.getLogger(__name__)


@job(VARIANT_CHANNEL, timeout=3 * 60 * 60)
def load_variants(manifest_path, database, **kwargs):
    "Variant loading requires only a VCF file and will never load a duplicate."
    manifest = ManifestReader(manifest_path)

    vcf_info = manifest.section('vcf')

    # No data regarding VCF
    if 'file' not in vcf_info:
        return

    cursor = connections[database].cursor()

    vcf_path = os.path.join(os.path.dirname(manifest_path), vcf_info['file'])

    with open(vcf_path) as fin:
        log.debug("opening {0} in {1}".format(vcf_path, __name__))
        stream = VariantStream(fin)
        columns = stream.output_columns
        db_table = Variant._meta.db_table
        pgcopy_batch(stream, db_table, columns, cursor, database)

    VARIANT_CHANNEL.publish(manifest_path=manifest_path, database=database)


@job(timeout=3 * 60 * 60)
def load_effects(manifest_path, database, **kwargs):
    manifest = ManifestReader(manifest_path)

    vcf_info = manifest.section('vcf')

    # No data regarding VCF
    if 'file' not in vcf_info:
        return

    cursor = connections[database].cursor()

    vcf_path = os.path.join(os.path.dirname(manifest_path), vcf_info['file'])

    with open(vcf_path) as fin:
        log.debug("opening {0} in {1} load_effects".format(vcf_path, __name__))
        stream = EffectStream(fin)
        columns = stream.output_columns
        db_table = VariantEffect._meta.db_table
        pgcopy_batch(stream, db_table, columns, cursor, database)


@job
def load_1000g(database, **kwargs):
    if not db.utils.table_exists('1000g', schema='raw'):
        return

    cursor = connections[database].cursor()
    cursor.execute(db.utils.sequence_reset_sql(ThousandG, database))

    with transaction.commit_manually(database):
        try:
            # Get all missing records..
            cursor.execute('''
                INSERT INTO "1000g" (
                    "variant_id", "an", "ac", "af", "aa",
                    "amr_af", "asn_af", "afr_af", "eur_af"
                ) (
                    SELECT variant.id, r.an, r.ac, r.af, r.aa, r.amr_af,
                        r.asn_af, r.afr_af, r.eur_af
                    FROM "variant"
                        INNER JOIN "raw"."1000g" r ON ("variant".md5 = r."md5")
                        LEFT OUTER JOIN "1000g"
                            ON ("1000g"."variant_id" = "variant"."id")
                    WHERE "1000g"."id" IS NULL
                )
            ''')
            transaction.commit()
        except Exception as e:
            transaction.rollback()
            log.exception(e)


@job
def load_sift(database, **kwargs):
    if not db.utils.table_exists('sift', schema='raw'):
        return

    cursor = connections[database].cursor()
    cursor.execute(db.utils.sequence_reset_sql(Sift, database))

    with transaction.commit_manually(database):
        try:
            cursor.execute('''
                INSERT INTO "sift" (
                    "variant_id", "score", "refaa", "varaa"
                ) (
                    SELECT variant.id, r.score, r.refaa, r.varaa
                    FROM "variant"
                        INNER JOIN "raw"."sift" r ON ("variant".md5 = r."md5")
                        LEFT OUTER JOIN "sift"
                            ON ("sift"."variant_id" = "variant"."id")
                    WHERE "sift"."id" IS NULL
                )
            ''')
            transaction.commit()
        except Exception as e:
            transaction.rollback()
            log.exception(e)


@job
def load_polyphen2(database, **kwargs):
    if not db.utils.table_exists('polyphen2', schema='raw'):
        return

    cursor = connections[database].cursor()
    cursor.execute(db.utils.sequence_reset_sql(PolyPhen2, database))

    with transaction.commit_manually(database):
        try:
            cursor.execute('''
                INSERT INTO "polyphen2" (
                    "variant_id", "score", "refaa"
                ) (
                    SELECT variant.id, r.score, r.refaa
                    FROM "variant"
                        INNER JOIN "raw"."polyphen2" r
                            ON ("variant".md5 = r."md5")
                        LEFT OUTER JOIN "polyphen2"
                            ON ("polyphen2"."variant_id" = "variant"."id")
                    WHERE "polyphen2"."id" IS NULL
                )
            ''')
            transaction.commit()
        except Exception as e:
            transaction.rollback()
            log.exception(e)


@job
def load_evs(database, **kwargs):
    if not db.utils.table_exists('evs', schema='raw'):
        return

    cursor = connections[database].cursor()
    cursor.execute(db.utils.sequence_reset_sql(EVS, database))

    with transaction.commit_manually(database):
        try:
            # MAFs are divided by 100 since they are percentages to begin with
            cursor.execute('''
                SELECT variant.id, r.ea_ac_ref, r.ea_ac_alt, r.aa_ac_ref,
                    r.aa_ac_alt, r.all_ac_ref, r.all_ac_alt, r.ea_maf / 100.0,
                    r.aa_maf / 100.0, r.all_maf / 100.0, r.gts, r.ea_gtc,
                    r.aa_gtc, r.all_gtc, r.clinical_association
                FROM "variant"
                    INNER JOIN "raw"."evs" r ON ("variant".md5 = r."md5")
                    LEFT OUTER JOIN "evs"
                        ON ("evs"."variant_id" = "variant"."id")
                WHERE "evs"."id" IS NULL
            ''')

            # Note, *_af are actually minor allele frequencies by default.
            # The handler checks to see if the reference count is less than
            # the alternate and sets the AF if the reference count is less than
            # the alternate
            columns = ['variant_id', 'ea_ac_ref', 'ea_ac_alt', 'aa_ac_ref',
                       'aa_ac_alt', 'all_ac_ref', 'all_ac_alt', 'ea_af',
                       'aa_af', 'all_af', 'gts', 'ea_gtc', 'aa_gtc', 'all_gtc',
                       'clinical_association']

            def compare_counts(ref, alt):
                "Compares allele counts and handles heterozygotes."
                ref = int(ref)
                for _alt in alt.split(','):
                    if ref < int(_alt):
                        return True
                return False

            def handler(row):
                record = OrderedDict(zip(columns, row))

                # All
                if compare_counts(record['all_ac_ref'], record['all_ac_alt']):
                    record['all_af'] = 1 - record['all_af']

                # European
                if compare_counts(record['ea_ac_ref'], record['ea_ac_alt']):
                    record['ea_af'] = 1 - record['ea_af']

                # African American
                if compare_counts(record['aa_ac_ref'], record['aa_ac_alt']):
                    record['aa_af'] = 1 - record['aa_af']

                cleaned = [str(x) if x is not None else '\N' for x in
                           record.values()]
                return '\t'.join(cleaned) + '\n'

            def streamer(cursor):
                while True:
                    rows = cursor.fetchmany(100)
                    if not rows:
                        break

                    for row in rows:
                        yield handler(row)

            pgcopy_batch(streamer(cursor), EVS._meta.db_table, columns=columns,
                         database=database)

            transaction.commit()
        except Exception as e:
            transaction.rollback()
            log.exception(e)



import hashlib


def record_is_valid(record):
    "Checks if a record is valid for processing."

    # No random contigs
    if record.CHROM.startswith('GL'):
        return False

    # Skip results with a read depth < 5. If no read depth is specified then
    # we have no choice but to consider this record as being valid.
    if 'DP' in record.INFO and record.INFO['DP'] < 5:
        return False

    return True


def file_md5(filename, block_size=2**16):
    md5 = hashlib.md5()
    with open(filename, 'rb') as fin:
        while True:
            data = fin.read(block_size)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()



import logging
import time
from optparse import make_option
from django.db import transaction, connections, DEFAULT_DB_ALIAS, DatabaseError
from django.db.models import F, Q
from django.core.management.base import BaseCommand
from vdw.samples.models import Cohort

log = logging.getLogger(__name__)


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Specifies the target database to load results.'),
        make_option('--force', action='store_true', dest='force',
                    default=False,
                    help='Forces recomputation of all cohort allele '
                         'frequencies')
    )

    def handle(self, **options):
        database = options.get('database')
        cursor = connections[database].cursor()

        if options.get('force'):
            cohorts = list(Cohort.objects.all())
        else:
            cohorts = list(Cohort.objects.filter(
                Q(allele_freq_modified=None) |
                Q(modified__gt=F('allele_freq_modified'))))

        log.debug('Computing for {0} cohorts'.format(len(cohorts)))

        with transaction.commit_manually(database):
            try:
                cursor.execute('TRUNCATE "cohort_variant" RESTART IDENTITY')
                transaction.commit()
            except DatabaseError:
                transaction.rollback()
                log.exception('Could not truncate CohortVariant table')

        for cohort in cohorts:
            log.debug('"{0}" ({1} samples)...'.format(cohort, cohort.count))

            t0 = time.time()
            with transaction.commit_manually(database):
                try:
                    cohort.compute_allele_frequencies(database)
                    log.debug('done in {0}s\n'.format(int(time.time() - t0)))
                    transaction.commit()
                except DatabaseError, e:
                    transaction.rollback()
                    log.exception(e)



import json
import logging
import requests
from datetime import datetime
from django.conf import settings
from optparse import make_option
from django.db import transaction, DEFAULT_DB_ALIAS
from django.core.management.base import BaseCommand
from requests.exceptions import SSLError, ConnectionError, RequestException
from vdw.samples.models import Sample, ResultScore

log = logging.getLogger(__name__)


class Command(BaseCommand):
    args = '<sample_label sample_label ...>'
    option_list = BaseCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Specifies the target database to load results.'),
        make_option('--force', action='store_true', dest='force',
                    default=False,
                    help='Forces recomputation of all gene rankings')
    )

    def handle(self, *args, **options):
        if not getattr(settings, 'PHENOTYPE_ENDPOINT', None):
            log.error('PHENOTYPE_ENDPOINT must be defined in settings for '
                      'gene rankings to be updated.')
            return

        if not getattr(settings, 'GENE_RANK_BASE_URL', None):
            log.error('GENE_RANK_BASE_URL must be defined in settings for '
                      'gene rankings to be updated.')
            return

        if (not getattr(settings, 'VARIFY_CERT', None) or
                not getattr(settings, 'VARIFY_KEY', None)):
            log.error('VARIFY_CERT and VARIFY_KEY must be defined in settings '
                      'for gene rankings to be updated.')
            return

        database = options.get('database')
        force = options.get('force')

        # Construct the cert from the setting to use in requests to the
        # phenotype endpoint.
        cert = (settings.VARIFY_CERT, settings.VARIFY_KEY)

        # We ignore all the samples that aren't published. They aren't visible
        # to the user so we don't bother updating related scores. If there
        # were sample labels supplied as arguments then we limit the rankings
        # updates to those samples, otherwise we process all samples.
        samples = Sample.objects.filter(published=True)

        if args:
            samples = samples.filter(label__in=args)

        updated_samples = 0
        total_samples = 0

        for sample in samples:
            total_samples += 1

            # Construct the URL from the setting and the sample label. The
            # sample label is used to retrieve the phenotype info on the remote
            # endpoint.
            url = settings.PHENOTYPE_ENDPOINT.format(sample.label)

            # Get the phenotype information for this sample. If the
            # phenotype is unavailable then we can skip this sample.
            try:
                response = requests.get(url, cert=cert, verify=False)
            except SSLError:
                log.exception('Skipping sample "{0}". An SSLError occurred '
                              'during phenotype retrieval request.'
                              .format(sample.label))
                continue
            except ConnectionError:
                log.exception('Skipping sample "{0}". A ConnectionError '
                              'occurred during phenotype retrieval request.'
                              .format(sample.label))
                continue
            except RequestException:
                log.exception('Skipping sample "{0}". The sample has no '
                              'phenotype data associated with it'
                              .format(sample.label))
                continue

            try:
                phenotype_data = json.loads(response.content)
            except ValueError:
                log.warning(
                    "Could not parse response from {0}, skipping '{1}'."
                    .format(url, sample.label))
                continue

            try:
                phenotype_modified = datetime.strptime(
                    phenotype_data['last_modified'] or '',
                    "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                phenotype_modified = datetime.min
                log.warn("Could not parse 'last_modified' field on phenotype "
                         "data. Using datetime.min so that only unranked "
                         "samples will be ranked. If the 'force' flag was "
                         "used then all samples will be updated despite this "
                         "parsing failure.")

            # If the parsed response doesn't contain any HPO terms then we can
            # skip this sample since we cannot rank genes without HPO terms.
            if not phenotype_data.get('hpoAnnotations'):
                log.warning("Response from phenotype missing HPO Annotations, "
                            "skipping '{0}'.".format(sample.label))
                continue

            if (not force and sample.phenotype_modified and
                    sample.phenotype_modified > phenotype_modified):
                log.debug("Sample '{0}' is already up to date, skipping it."
                          .format(sample.label))
                continue

            # Extract the HPO terms from the data returned from the phenotype
            # endpoint. We need to modify the terms a bit because the phenotype
            # endpoint has terms in the form 'HP_0011263' and the gene ranking
            # enpoint expects them to be of the form 'HP:0011263'.
            hpo_terms = []
            for hpo_annotation in phenotype_data['hpoAnnotations']:
                try:
                    hpo_id = hpo_annotation.get('hpo_id')
                except AttributeError:
                    continue

                if hpo_id:
                    hpo_terms.append(hpo_id.replace('_', ':'))

            # If there are no HPO terms then there will be no rankings so skip
            # this sample to avoid any more computations and requests.
            if not hpo_terms:
                log.warning('Skipping "{0}" because it has no HPO terms '
                            'associated with it.'.format(sample.label))
                continue

            # Compute the unique gene list for the entire sample
            genes = set(sample.results.values_list(
                'variant__effects__transcript__gene__symbol', flat=True))

            # Obviously, if there are no genes then the gene ranking endpoint
            # will have nothing to do so we can safely skip this sample.
            if not genes:
                log.warning('Skipping "{0}" because it has no genes '
                            'associated with it.'.format(sample.label))
                continue

            # Convert genes to a list so it is serializeable in the json.dumps
            # call below when making the request to the ranking service.
            data = {
                'hpo': hpo_terms,
                'genes': [g for g in genes if g]
            }

            try:
                gene_response = requests.post(
                    settings.GENE_RANK_BASE_URL, data=json.dumps(data),
                    headers={'content-type': 'application/json'})
            except Exception:
                log.exception('Error retrieving gene rankings, skipping '
                              'sample "{0}".'.format(sample.label))
                continue

            try:
                gene_data = json.loads(gene_response.content)
            except ValueError:
                log.warning(
                    "Could not parse response from {0}, skipping '{1}'."
                    .format(settings.GENE_RANK_BASE_URL, sample.label))
                continue

            ranked_genes = gene_data['ranked_genes']

            updated_results = 0
            total_results = 0

            for result in sample.results.all():
                total_results += 1

                with transaction.commit_manually(database):
                    try:
                        # Instead of trying to remove None from the returned
                        # values list we just exclude them from the query
                        # itself.
                        genes = result.variant.effects\
                            .exclude(transcript__gene__symbol__isnull=True)\
                            .order_by('effect__impact')\
                            .values_list(
                                'transcript__gene__symbol', flat=True)\
                            .distinct()

                        # If there is no gene on this result or the gene is
                        # not found in the list of ranked genes then skip this
                        # result.
                        if not genes:
                            log.debug("Result with id {0} has no gene, "
                                      "skipping result.".format(result.id))
                            transaction.rollback()
                            continue

                        # Use the first gene from the list since a result can
                        # have more than one gene associated with it, we
                        # return the first gene symbol in the list. This is
                        # the same one that will be shown in the collapsed
                        # gene list on the variant row in the results table.
                        gene = genes[0]

                        # Get the first item in the ranked gene list with a
                        # symbol matching the gene we looked up above for this
                        # result.
                        ranked_gene = None

                        for ranked_gene in ranked_genes:
                            if (ranked_gene.get('symbol', '').lower() ==
                                    gene.lower()):
                                break
                            else:
                                ranked_gene = None

                        if not ranked_gene:
                            log.debug("Could not find '{0}' in ranked gene "
                                      "list, skipping result".format(gene))
                            transaction.rollback()
                            continue

                        try:
                            rs = ResultScore.objects.get(result=result)
                            rs.rank = ranked_gene.get('rank')
                            rs.score = ranked_gene.get('score')
                        except ResultScore.DoesNotExist:
                            rs = ResultScore(
                                result=result,
                                rank=ranked_gene.get('rank'),
                                score=ranked_gene.get('score'))
                        rs.save()

                        updated_results += 1

                    except Exception:
                        log.exception("Error saving gene ranks and scores for "
                                      "sample '{0}'".format(sample.label))
                        transaction.rollback()
                        continue

                    transaction.commit()

            sample.phenotype_modified = datetime.now()
            sample.save()

            log.info("Updated {0} and skipped {1} results in sample '{2}'"
                     .format(updated_results, total_results - updated_results,
                             sample.label))
            updated_samples += 1

        log.info("Updated {0} and skipped {1} samples"
                 .format(updated_samples, total_samples-updated_samples))



from vdw.raw.management.subcommands.load import LoadCommand
from vdw.raw.utils.stream import VCFPGCopyEditor
from vdw.variants.utils import calculate_md5


class EVSProcessor(VCFPGCopyEditor):
    output_columns = ('chr', 'pos', 'ref', 'alt', 'rsid', 'ea_ac_alt',
                      'ea_ac_ref', 'aa_ac_alt', 'aa_ac_ref', 'all_ac_alt',
                      'all_ac_ref', 'ea_maf', 'aa_maf', 'all_maf', 'gts',
                      'ea_gtc', 'aa_gtc', 'all_gtc', 'read_depth',
                      'clinical_association', 'md5')

    vcf_fields = ('CHROM', 'POS', 'REF', 'ALT', 'ID')

    info_fields = ('EA_AC', 'AA_AC', 'TAC', 'MAF', 'GTS', 'EA_GTC', 'AA_GTC',
                   'GTC', 'DP', 'CA')

    def process_column(self, key, value):
        # *AC => allele counts. The format is "alt,ref"
        if key in ('EA_AC', 'AA_AC', 'TAC'):
            value = value.split(',')
            value = [','.join(value[:-1]), value[-1]]
            return value
        # MAF => allele frequencies. The format is 'EA,AA,All'
        if key == 'MAF':
            return value.split(',')
        return super(EVSProcessor, self).process_column(key, value)

    def process_line(self, record):
        cleaned = super(EVSProcessor, self).process_line(record)
        # Add the MD5
        cleaned.append(calculate_md5(*cleaned[:4]))
        return cleaned


class Command(LoadCommand):
    """Loader for raw EVS data.

    Website: http://evs.gs.washington.edu/EVS/
    """
    targets = ['evs']

    qualified_names = {
        'evs': '"raw"."evs"',
    }

    create_sql = {
        'evs': '''CREATE TABLE {} (
            "chr" text NOT NULL,
            "pos" integer NOT NULL,
            "alt" text NOT NULL,
            "ref" text NOT NULL,
            "md5" varchar(32) NOT NULL,
            "rsid" varchar(30),
            "ea_ac_ref" varchar(20),
            "ea_ac_alt" varchar(20),
            "aa_ac_ref" varchar(20),
            "aa_ac_alt" varchar(20),
            "all_ac_ref" varchar(20),
            "all_ac_alt" varchar(20),
            "ea_maf" double precision,
            "aa_maf" double precision,
            "all_maf" double precision,
            "gts" varchar(200),
            "ea_gtc" varchar(200),
            "aa_gtc" varchar(200),
            "all_gtc" varchar(200),
            "read_depth" integer,
            "clinical_association" text
        )''',
    }

    processors = {
        'evs': EVSProcessor,
    }



from django.db import models
from django.contrib.auth.models import User
from vdw.core.models import BigForeignKey, TimestampedModel
from vdw.samples.models import Result
import reversion


class Pathogenicity(models.Model):
    name = models.CharField(max_length=200, unique=True)

    class Meta(object):
        db_table = 'pathogenicity'
        verbose_name_plural = 'pathogenicities'

    def __unicode__(self):
        return self.name


class AssessmentCategory(models.Model):
    name = models.CharField(max_length=200, unique=True)

    class Meta(object):
        db_table = 'assessment_category'
        verbose_name_plural = 'assessment categories'

    def __unicode__(self):
        return self.name


class ParentalResult(models.Model):
    name = models.CharField(max_length=200, unique=True)

    class Meta(object):
        db_table = 'parental_result'

    def __unicode__(self):
        return self.name


class SangerResult(models.Model):
    confirmed = models.BooleanField()

    class Meta(object):
        db_table = 'sanger_result'


class Assessment(TimestampedModel):
    evidence_details = models.TextField(null=True, blank=True)
    sanger_requested = models.BooleanField()

    assessment_category = models.ForeignKey(AssessmentCategory)
    father_result = models.ForeignKey(ParentalResult, related_name='father')
    mother_result = models.ForeignKey(ParentalResult, related_name='mother')
    pathogenicity = models.ForeignKey(Pathogenicity)
    sample_result = BigForeignKey(Result)

    sanger_result = models.ForeignKey(SangerResult, null=True, blank=True)
    user = models.ForeignKey(User)

    class Meta(object):
        db_table = 'assessment'
        unique_together = ('sample_result', 'user')


reversion.register(Pathogenicity)
reversion.register(AssessmentCategory)
reversion.register(ParentalResult)
reversion.register(SangerResult)
reversion.register(Assessment,
                   follow=["assessment_category", "father_result",
                           "mother_result", "pathogenicity", "sanger_result"])



import hashlib
from vdw.raw.management.subcommands.load import LoadCommand
from vdw.raw.utils.stream import PGCopyEditor


class SIFTProcessor(PGCopyEditor):
    def __init__(self, *args, **kwargs):
        kwargs['indices'] = (0, 1, 3, 4, 5, 6, 7)
        super(SIFTProcessor, self).__init__(*args, **kwargs)

    def process_column(self, idx, value):
        # Incrementally build up the tokens for calculating the MD5
        if idx == 0:
            self.md5_toks = [value]
        if idx in (1, 3):
            self.md5_toks.append(value)
        if idx == 4:
            self.md5_toks.append(value)
            return [value, hashlib.md5('|'.join(self.md5_toks)).hexdigest()]
        return super(SIFTProcessor, self).process_column(idx, value)


class Command(LoadCommand):
    targets = ['sift']

    qualified_names = {
        'sift': '"raw"."sift"',
    }

    drop_sql = {
        'sift': 'DROP TABLE IF EXISTS {}',
    }

    create_sql = {
        'sift': '''CREATE TABLE {} (
            "chr" text NOT NULL,
            "pos" integer NOT NULL,
            "ref" text NOT NULL,
            "alt" text NOT NULL,
            "md5" varchar(32) NOT NULL,
            "score" double precision,
            "refaa" varchar(2),
            "varaa" varchar(2)
        )''',
    }

    processors = {
        'sift': SIFTProcessor,
    }



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'PubMed'
        db.create_table('pubmed', (
            ('pmid', self.gf('django.db.models.fields.IntegerField')(primary_key=True)),
        ))
        db.send_create_signal('literature', ['PubMed'])


    def backwards(self, orm):
        
        # Deleting model 'PubMed'
        db.delete_table('pubmed')


    models = {
        'literature.pubmed': {
            'Meta': {'object_name': 'PubMed', 'db_table': "'pubmed'"},
            'pmid': ('django.db.models.fields.IntegerField', [], {'primary_key': 'True'})
        }
    }

    complete_apps = ['literature']



from django.core import exceptions
from django.db import models
from django.db.models import fields
from south.modelsinspector import add_introspection_rules


class TimestampedModel(models.Model):
    notes = models.TextField(null=True, blank=True)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta(object):
        abstract = True


class LabeledModel(models.Model):
    """A labeled model is simply for differentiating between an editable
    human-readable label and some unique name that should not change. This
    is necessary if other processes are dependent on the name.
    """
    # This is the label for display that can be freely changed without
    # affecting downstream processes such as loading
    label = models.CharField(max_length=100)

    # The fixed name of the project that other processes can depend on
    # such as lookups while loading samples
    name = models.CharField(max_length=100, editable=False)

    class Meta(object):
        abstract = True

    def __unicode__(self):
        return self.label


class BigIntegerField(fields.IntegerField):
    def db_type(self, connection):
        if 'postgres' in connection.__class__.__module__:
            return 'bigint'
        else:
            raise NotImplemented

    def get_internal_type(self):
        return 'BigIntegerField'

    def to_python(self, value):
        if value is None:
            return value
        try:
            return long(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                'This value must be a long integer.')


class BigAutoField(fields.AutoField):
    # ONLY PostgreSQL is supported (MySQL, Oracle & SQLite are NOT supported)
    def db_type(self, connection):
        if 'postgres' in connection.__class__.__module__:
            return 'bigserial'
        else:
            raise NotImplemented

    def get_internal_type(self):
        return 'BigAutoField'

    def to_python(self, value):
        if value is None:
            return value
        try:
            return long(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                'This value must be a long integer.')


class BigForeignKey(fields.related.ForeignKey):
    def db_type(self, connection):
        rel_field = self.rel.get_related_field()
        is_bigauto = isinstance(rel_field, BigAutoField)
        is_bigint = isinstance(rel_field, BigIntegerField)
        matches_type = connection.features.related_fields_match_type
        if (is_bigauto or (not matches_type and is_bigint)):

            return BigIntegerField().db_type(connection)
        return rel_field.db_type(connection)


add_introspection_rules([], ["^vdw\.core\.models\.BigIntegerField"])
add_introspection_rules([], ["^vdw\.core\.models\.BigAutoField"])
add_introspection_rules([], ["^vdw\.core\.models\.BigForeignKey"])



from django.db import models


class PubMed(models.Model):
    pmid = models.IntegerField(primary_key=True)

    class Meta(object):
        db_table = 'pubmed'

    @property
    def pmid_url(self):
        return 'http://www.ncbi.nlm.nih.gov/pubmed/' + str(self.pmid)



import re
from optparse import make_option
from django.db import models, QName
from django.db import connections, DEFAULT_DB_ALIAS
from django.core.management.base import AppCommand
from django.core.management.color import no_style


def sql_drop_indexes_for_field(model, f, style, connection):
    """
    Return the DROP INDEX SQL statements for a single model field.
    """
    output = []
    cqn = connection.ops.compose_qualified_name
    qname = connection.qualified_name(model)

    # HACK
    create_indexes = \
        connection.creation.sql_indexes_for_field(model, f, no_style())
    for statement in create_indexes:
        name = re.search(r'CREATE INDEX (.*) ON', statement).groups()[0]
        name = cqn(
            QName(schema=qname.schema, table=name, db_format=qname.db_format))
        output.append(style.SQL_KEYWORD(
            'DROP INDEX IF EXISTS') + ' ' + style.SQL_TABLE(name) + ';')
    return output


def sql_drop_indexes_for_model(model, style, connection):
    """
    Returns the DROP INDEX SQL statements for a single model.
    """
    if not model._meta.managed or model._meta.proxy:
        return []
    output = []
    for f in model._meta.local_fields:
        output.extend(sql_drop_indexes_for_field(model, f, style, connection))
    return output


def sql_drop_indexes(app, style, connection):
    output = []
    for model in models.get_models(app):
        output.extend(sql_drop_indexes_for_model(model, style, connection))
    return output


class Command(AppCommand):
    help = ("Prints the DROP INDEX SQL statements for the given app module "
            "name(s).")

    option_list = AppCommand.option_list + (
        make_option('--database', action='store', dest='database',
                    default=DEFAULT_DB_ALIAS,
                    help='Nominates a database to print the SQL for. Defaults '
                         'to the "default" database.'),
    )

    output_transaction = True

    def handle_app(self, app, **options):
        return u'\n'.join(sql_drop_indexes(
            app,
            self.style,
            connections[options.get('database')])).encode('utf-8')



import logging
from django.contrib.auth.models import User
from django.conf import settings
from django.db import router, models, transaction, connections, DatabaseError
from sts.contextmanagers import transition
from objectset.models import ObjectSet, SetObject
from vdw.core.models import BigAutoField, BigForeignKey, LabeledModel, \
    TimestampedModel
from vdw.genome.models import Genotype
from vdw.variants.models import Variant

# The catch all project. If this is None, samples will not be added to
# a default project.
DEFAULT_PROJECT_NAME = getattr(settings, 'DEFAULT_PROJECT_NAME', 'Default')

# Cohort of the entire population of samples. If this is None, samples
# will not be added to a default cohort
DEFAULT_COHORT_NAME = getattr(settings, 'DEFAULT_COHORT_NAME', 'World')

log = logging.getLogger(__name__)


class Person(TimestampedModel):
    "A single person, either the proband or a relative."
    mrn = models.CharField(max_length=50, null=True, blank=True)

    # For pedigree and sex-linked analyses
    sex = models.CharField(max_length=20, null=True, blank=True)

    # Designates whether this person is the proband, i.e. person of
    # interest in the context of variant analysis, e.g. family members
    # are not.
    proband = models.BooleanField(default=False)

    # For trio or general pedigree analysis, this marks an explicit link
    # between the proband and the family member.
    relations = models.ManyToManyField('self', through='Relation',
                                       symmetrical=False)

    class Meta(object):
        db_table = 'person'
        permissions = (
            ('view_person', 'View Person'),
        )

    def __unicode__(self):
        return self.mrn or 'P{0}'.format(str(self.pk).zfill(6))


class Relation(TimestampedModel):
    "A relation between two persons. This enables pedigree analysis."
    person = models.ForeignKey(Person, related_name='family')
    relative = models.ForeignKey(Person, related_name='relative_of')

    # Type of relation, e.g. mother, father, sibling
    type = models.CharField(max_length=20)

    # Generation of `relative` relative to `person` where the generation
    # of `person` is 0, e.g. a parent is 1 and a sibling is 0.
    generation = models.IntegerField(default=0)

    class Meta(object):
        db_table = 'relation'
        ordering = ('person', '-generation')


class Project(LabeledModel, TimestampedModel):
    "High-level organization of cohorts (and samples) within a project."

    class Meta(LabeledModel.Meta):
        db_table = 'project'
        unique_together = ('name',)
        permissions = (
            ('view_project', 'View Project'),
        )


class Batch(LabeledModel, TimestampedModel):
    """A cohort defines a high-level grouping of samples for a particular
    study and/or research question. This is referred to as project-level
    cohort.
    """
    investigator = models.CharField(max_length=100, null=True, blank=True)

    # The project this cohort (and thus samples) are associated with)
    project = models.ForeignKey(Project, related_name='batches')

    # Defines whether this cohort is published. Once at least one sample
    # if loaded, this cohort may become available.
    published = models.BooleanField(default=False)

    # Store the sample count since this does not change
    count = models.IntegerField(default=0)

    class Meta(LabeledModel.Meta):
        db_table = 'batch'
        unique_together = ('project', 'name')
        ordering = ('project', 'label')
        verbose_name_plural = 'batches'
        permissions = (
            ('view_batch', 'View Batch'),
        )


class Sample(LabeledModel, TimestampedModel):
    # Direct reference to project this sample is associated with. Although
    # this is available through `batch`, batches are a function of the loading
    # process, but not necessarily the hierarchy of samples.
    project = models.ForeignKey(Project, related_name='samples')

    # Reference to the cohort this sample is associated with
    batch = models.ForeignKey(Batch, related_name='samples')

    # Version of the sample and downstream results
    version = models.IntegerField()

    # Reference to the person this sample is derived from.
    person = models.ForeignKey(Person, null=True, blank=True,
                               related_name='samples')

    # Store the result count since this does not change
    count = models.IntegerField(default=0)

    # Identifier of the biological sample this data is associated with
    # in the LIMS
    bio_sample = models.IntegerField(null=True, blank=True)

    # The name this sample was given in the VCF file (as opposed to the
    # MANIFEST)
    vcf_colname = models.CharField(max_length=200, null=True, blank=True,
                                   help_text='vcf column sample name')

    # Marks whether this sample is published. Once this sample's results have
    # been fully processed will this flag be true. As a sanity check, there
    # should only ever be one sample published for a single person.
    published = models.BooleanField(default=False)

    # MD5 of the VCF sample file. This provides a sanity check for
    # new files that represent the same sample in case reloads need
    # to happen.
    md5 = models.CharField(max_length=32, null=True, blank=True,
                           editable=False)

    # This indicates the last time(if any) that this sample had its gene
    # rankings updated using its associated phenotype terms. Since there is
    # no guarantee that this sample has been ranked or that it even has
    # phenotype data, we allow this field to be null.
    phenotype_modified = models.DateTimeField(null=True, blank=True)

    class Meta(LabeledModel.Meta):
        db_table = 'sample'
        unique_together = ('batch', 'name', 'version')
        ordering = ('project', 'batch', 'label')
        permissions = (
            ('view_sample', 'View Sample'),
        )

    def __unicode__(self):
        return self.label


class SampleManifest(TimestampedModel):
    "Discrete attributes from the sample manifest."
    sample = models.OneToOneField(Sample, related_name='manifest')
    # Original path to the manifest file manifest paths cannot be unique
    # since a vcf can contain more than one vcf.
    path = models.CharField(max_length=200, unique=False)
    # Cache of the manifest content in case the file no longer exists
    content = models.TextField()

    class Meta(object):
        db_table = 'sample_manifest'

    def __unicode__(self):
        return u'Sample Manifest: {0}'.format(unicode(self.sample))

    def load_content(self, path):
        "Attempts to open the file and read the content."
        log.debug('Opening {0} in {1} load_content'.format(path, __name__))
        with open(path) as fin:
            self.path = path
            self.content = fin.read()

    def content_has_changed(self):
        "Reads the contents from the file and compares with the local copy."

        log.debug('Opening {0} in {1} content_has_changed'
                  .format(self.path, __name__))
        with open(self.path) as fin:
            return fin.read() != self.content


class Cohort(ObjectSet):
    "A Cohort is a logical grouping of samples for some purpose."

    user = models.ForeignKey(User, null=True, blank=True)
    name = models.CharField(max_length=100, null=True, blank=True)

    # Reference for easier management and delegation of permissions
    # for related users
    project = models.ForeignKey(Project, null=True, blank=True)

    # Reference for easier autocreation
    batch = models.ForeignKey(Batch, null=True, blank=True)

    samples = models.ManyToManyField(Sample, through='CohortSample')

    # Defines whether this cohort is published. This can be used
    # to toggle when this cohort is visible depending on when if
    # the statistics across samples is consistent.
    published = models.BooleanField(default=False)

    autocreated = models.BooleanField(default=False)

    # This is purely for ordering the autocreated cohorts above
    # user-defined cohorts..
    order = models.SmallIntegerField(default=0)

    allele_freq_modified = models.DateTimeField(null=True)

    set_object_rel = 'samples'
    label_field = 'name'

    def __unicode__(self):
        return unicode(self.name)

    class Meta(object):
        db_table = 'cohort'
        ordering = ('-order', 'name')
        permissions = (
            ('view_cohort', 'View Cohort'),
        )

    def compute_allele_frequencies(self, using=None):
        "Computes the allele frequencies across all samples in this cohort."
        if using is None:
            using = router.db_for_write(self.__class__)
        cursor = connections[using].cursor()

        with transition(self, 'Recomputed Allele Frequencies'):
            with transaction.commit_manually(using):
                # TODO update to use model._meta.db_table..
                try:
                    # Update count on cohort instance
                    cursor.execute('''
                        UPDATE "cohort" SET "count" = (
                            SELECT COUNT(id)
                            FROM "cohort_sample"
                            WHERE "cohort_id" = "cohort"."id"
                        ) WHERE "id" = %s
                    ''', [self.id])

                    # Calculate frequencies for all variants associated with
                    # all samples in this cohort
                    cursor.execute('''
                        INSERT INTO cohort_variant (cohort_id, variant_id, af)
                        (
                            SELECT c.id, r.variant_id,
                                COUNT(r.id) / c."count"::float
                            FROM sample_result r
                                INNER JOIN sample s ON (r.sample_id = s.id)
                                INNER JOIN cohort_sample cs ON
                                    (cs.sample_id = s.id)
                                INNER JOIN cohort c ON (cs.cohort_id = c.id)
                            WHERE c.id = %s
                            GROUP BY c.id, r.variant_id, c."count"
                        )
                    ''', [self.id])
                    transaction.commit()
                except DatabaseError, e:
                    transaction.rollback()
                    log.exception(e)
                    raise


class CohortSample(SetObject):
    object_set = models.ForeignKey(Cohort, db_column='cohort_id')
    set_object = models.ForeignKey(Sample, db_column='sample_id')

    class Meta(SetObject.Meta):
        db_table = 'cohort_sample'


class CohortVariant(models.Model):
    """Cohort-level aggregated data relative to specific variants. This
    currently only includes 'allele frequency' for the given cohort.
    """

    # Explictly declare BigAutoField primary key to avoid reaching max int
    id = BigAutoField(primary_key=True)

    variant = models.ForeignKey(Variant, related_name='cohort_details')
    cohort = models.ForeignKey(Cohort)
    af = models.FloatField(null=True, db_index=True)

    class Meta(object):
        db_table = 'cohort_variant'
        unique_together = ('variant', 'cohort')


class Result(TimestampedModel):
    # Explictly declare BigAutoField primary key to avoid reaching max int
    id = BigAutoField(primary_key=True)

    # Reference to the sample
    sample = models.ForeignKey(Sample, related_name='results')

    # Reference to the unique variant this result is about
    variant = models.ForeignKey(Variant)

    # Genome Analysis Toolkit (GATK) VCF fields
    quality = models.FloatField(null=True, blank=True, db_index=True)

    # DP
    read_depth = models.IntegerField(null=True, blank=True, db_index=True)
    raw_read_depth = models.IntegerField(null=True, blank=True, db_index=True)

    # GATK VCF genotype fields GT
    genotype = models.ForeignKey(Genotype, null=True, blank=True)
    genotype_quality = models.FloatField(null=True, blank=True)

    # Depth of coverage per allele
    coverage_ref = models.IntegerField(null=True, blank=True)
    coverage_alt = models.IntegerField(null=True, blank=True)

    phred_scaled_likelihood = models.TextField(null=True, blank=True)

    in_dbsnp = models.BooleanField(default=False, db_index=True)
    downsampling = models.NullBooleanField(blank=True)
    spanning_deletions = models.FloatField(null=True, blank=True)
    mq = models.FloatField('mapping quality', null=True, blank=True)
    mq0 = models.FloatField('mapping quality zero', null=True, blank=True)
    baseq_rank_sum = models.FloatField('base quality rank sum test',
                                       null=True, blank=True)
    mq_rank_sum = models.FloatField('mapping quality rank sum test',
                                    null=True, blank=True)
    read_pos_rank_sum = models.FloatField(null=True, blank=True)
    strand_bias = models.FloatField(null=True, blank=True)
    homopolymer_run = models.IntegerField(null=True, blank=True)
    haplotype_score = models.FloatField(null=True, blank=True)
    quality_by_depth = models.FloatField(null=True, blank=True)
    fisher_strand = models.FloatField(null=True, blank=True)
    base_counts = models.CharField(max_length=100, null=True, blank=True,
                                   help_text='A,C,G,T base counts')

    class Meta(object):
        db_table = 'sample_result'
        unique_together = ('sample', 'variant')

    @property
    def genotype_value(self):
        if self.genotype:
            genotype = self.genotype.value
            variant = self.variant
            if genotype in ('1/1', '1/2'):
                return variant.alt + '/' + variant.alt
            elif genotype == '0/1':
                return variant.ref + '/' + variant.alt
            elif genotype == '0/0':
                return variant.ref + '/' + variant.alt

    @property
    def base_count_map(self):
        if self.base_counts:
            bases = ('A', 'C', 'G', 'T')
            try:
                counts = [int(x) for x in self.base_counts.split(',')]
                return dict(zip(bases, counts))
            except (TypeError, ValueError) as e:
                log.exception(e)


class ResultScore(TimestampedModel):
    result = BigForeignKey(Result, related_name='score', unique=True)
    rank = models.IntegerField()
    score = models.FloatField()

    class Meta(object):
        db_table = 'result_score'


class ResultSet(ObjectSet):
    sample = models.ForeignKey(Sample, related_name='result_sets')
    name = models.CharField(max_length=200)
    description = models.TextField(null=True, blank=True)
    user = models.ForeignKey(User)

    results = models.ManyToManyField(Result, through='ResultSetItem')

    class Meta(object):
        db_table = 'result_set'
        unique_together = ('sample', 'name')


class ResultSetItem(SetObject):
    result = models.ForeignKey(Result)
    set = models.ForeignKey(ResultSet)

    class Meta(object):
        db_table = 'result_set_item'
        unique_together = ('result', 'set')


# Load signal receivers. This is imported below to prevent circular imports
# with the above models.
from . import receivers     # noqa



import sys
from setuptools import setup, find_packages

PACKAGE = 'vdw'
VERSION = __import__(PACKAGE).get_version()

install_requires = [
    'django>=1.4.11,<1.5',
    'south==0.8.2',
    'python-memcached==1.48',
    'coverage',
    'rq>=0.3.8',
    'django-rq>=0.5.1',
    'psycopg2==2.4.4',
    'avocado>=2.3.8,<3.0',
    'modeltree>=1.1.7',
    'django-objectset>=0.2.2',
    'django-sts==0.7.3',
    'django-reversion==1.6.6',
    'django-guardian==1.0.4',
    'diff-match-patch',
    'pyvcf>=0.6.5',
]

if sys.version_info < (2, 7):
    install_requires += ['importlib']

kwargs = {
    'name': PACKAGE,
    'version': VERSION,
    'packages': find_packages(exclude=['tests', '*.tests', '*.tests.*',
                                       'tests.*']),
    'include_package_data': True,
    'install_requires': install_requires,
    'test_suite': 'test_suite',
    'tests_require': ['httpretty'],
    'author': 'Byron Ruth',
    'author_email': 'b@devel.io',
    'description': 'Models and loader for processing, storing, and querying genetic data',  # noqa
    'license': 'BSD',
    'keywords': 'vcf varify harvest orm genome',
    'url': 'http://github.com/cbmi/varify-data-warehouse',
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
    ],
}

setup(**kwargs)



from vdw.pipeline import Channel

SAMPLE_CHANNEL = Channel('samples', ['manifest_path', 'database'])


# Setup local handlers to watch the channels of interest..
from vdw.variants.pipeline import VARIANT_CHANNEL
from . import handlers
handlers.load_results.watch(VARIANT_CHANNEL)



from django.db import models


class Source(models.Model):
    name = models.CharField(max_length=200)
    version = models.CharField(max_length=30, null=True, blank=True)
    release_date = models.DateField(null=True, blank=True)
    description = models.TextField(blank=True)
    website = models.TextField(blank=True)
    download_url = models.TextField(blank=True)

    published = models.BooleanField(default=False)
    archived = models.BooleanField(default=False)

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta(object):
        db_table = 'source'
        ordering = ['name']

    def __unicode__(self):
        if self.version:
            return u'{0} (v{1})'.format(self.name, self.version.lstrip('v'))
        if self.release_date:
            return u'{0} ({1})'.format(
                self.name, self.release_date.strftime('%Y-%m-%d'))
        return u'{0}'.format(self.name)


class Stat(models.Model):
    name = models.CharField(max_length=100)
    value = models.CharField(max_length=100)
    source = models.ForeignKey(Source, related_name='stats')

    class Meta(object):
        db_table = 'source_stat'



import logging
from datetime import datetime
from vdw.genome.models import Genotype
from vdw.variants.utils import calculate_md5
from vdw.variants.pipeline.utils import VariantCache
from vdw.pipeline import checks
from vdw.raw.utils.stream import VCFPGCopyEditor
from vdw.samples.models import Result

log = logging.getLogger(__name__)


class ResultStream(VCFPGCopyEditor):
    vcf_fields = ('CHROM', 'POS', 'REF', 'ALT')

    info_fields = ('DB', 'DS', 'DP', 'Dels', 'MQ', 'MQ0', 'BaseQRankSum',
                   'MQRankSum', 'ReadPosRankSum', 'SB', 'Hrun',
                   'HaplotypeScore', 'QD', 'FS', 'BaseCounts')

    output_columns = ('in_dbsnp', 'downsampling', 'raw_read_depth',
                      'spanning_deletions', 'mq', 'mq0', 'baseq_rank_sum',
                      'mq_rank_sum', 'read_pos_rank_sum', 'strand_bias',
                      'homopolymer_run', 'haplotype_score', 'quality_by_depth',
                      'fisher_strand', 'base_counts', 'variant_id',
                      'sample_id', 'read_depth', 'quality', 'genotype_id',
                      'genotype_quality', 'coverage_ref', 'coverage_alt',
                      'phred_scaled_likelihood', 'created', 'modified')

    def __init__(self, *args, **kwargs):
        self.sample_id = kwargs.pop('sample_id')
        self.vcf_sample = kwargs.pop('vcf_sample')
        self.now = datetime.now()
        super(ResultStream, self).__init__(*args, **kwargs)
        self.genotypes = dict(Genotype.objects.values_list('value', 'pk'))
        self.variants = VariantCache()

    def process_line(self, record):
        # we are revisiting the same line once for each sample
        # until I figure out how the cleaning and loading will
        # work otherwise

        # Ensures the cache is updated and available
        self.variants.ensure_cache(record)

        # Calculate the MD5 of the variant itself (not the record)
        md5 = calculate_md5(record)

        # Ensure the variant exists
        variant_id = self.variants.get(md5)
        assert variant_id is not None

        cleaned = super(ResultStream, self).process_line(record)
        # Remove variant specific parts
        cleaned = cleaned[4:]

        # in_dbsnp is not nullable so we need to replace null values with
        # False for this line to be valid. in_dbsnp is the first column as
        # determined by the `output_columns` defined above.
        if cleaned[0] == '\\N':
            cleaned[0] = False

        # can these be indexed?
        call = record.genotype(self.vcf_sample)

        # Already seen this variant for this sample, otherwise we would get a
        # duplicate key value violation in sample_result.
        if Result.objects.filter(
                variant=variant_id, sample=self.sample_id).exists():
            return None

        # The possibility for multiple alleles in these wide vcfs is almost
        # infinite so we need to triage the really weird ones into having a
        # reference allele "0/#" or being off the map entirely "#/#"".
        gt = getattr(call.data, 'GT', None)
        if gt:
            try:
                keyed_geno = self.genotypes[gt]
            except KeyError:
                if gt.startswith('0'):
                    keyed_geno = self.genotypes['0/#']
                else:
                    keyed_geno = self.genotypes['#/#']
        else:
            keyed_geno = None

        dp = getattr(call.data, 'DP', None)
        gq = getattr(call.data, 'GQ', None)
        ad = getattr(call.data, 'AD', None)
        if ad and len(ad) > 1:
            ad0 = ad[0]
            ad1 = ad[1]
        else:
            ad0 = None
            ad1 = None
        pl = getattr(call.data, 'PL', None)
        if pl:
            pl = ','.join([str(x) for x in pl])

        # Append remaining columns
        other = [variant_id, self.sample_id, dp, record.QUAL, keyed_geno, gq,
                 ad0, ad1, pl, self.now, self.now]

        cleaned.extend([self.process_column('', x) for x in other])
        return cleaned

    def readline(self, size=-1):
        "Ignore the `size` since a complete line must be processed."
        while True:
            try:
                record = next(self.reader)
            except StopIteration:
                break

            # Ensure this is a valid record
            if checks.record_is_valid(record):
                # Process the line
                cleaned = self.process_line(record)
                if cleaned:
                    return self.outdel.join([str(x) for x in cleaned]) + '\n'

        return ''



from vdw.raw.management.subcommands.load import LoadCommand
from vdw.raw.utils.stream import VCFPGCopyEditor
from vdw.variants.utils import calculate_md5


class ThousandGProcessor(VCFPGCopyEditor):
    vcf_fields = ('CHROM', 'POS', 'REF', 'ALT', 'ID')

    info_fields = ('AN', 'AC', 'AF', 'AA', 'AMR_AF', 'ASN_AF', 'AFR_AF',
                   'EUR_AF')

    # Output column names to match position
    output_columns = ('chr', 'pos', 'ref', 'alt', 'rsid', 'an', 'ac', 'af',
                      'aa', 'amr_af', 'asn_af', 'afr_af', 'eur_af', 'md5')

    def process_line(self, record):
        cleaned = super(ThousandGProcessor, self).process_line(record)
        # Add the MD5
        cleaned.append(calculate_md5(*cleaned[:4]))
        return cleaned


class Command(LoadCommand):
    targets = ['1000g']

    qualified_names = {
        '1000g': '"raw"."1000g"',
    }

    create_sql = {
        '1000g': '''CREATE TABLE {} (
            "chr" text NOT NULL,
            "pos" integer NOT NULL,
            "ref" text NOT NULL,
            "alt" text NOT NULL,
            "md5" varchar(32) NOT NULL,
            "rsid" varchar(30),
            "an" integer,
            "ac" integer,
            "af" double precision,
            "aa" text,
            "amr_af" double precision,
            "asn_af" double precision,
            "afr_af" double precision,
            "eur_af" double precision
        )''',
    }

    processors = {
        '1000g': ThousandGProcessor,
    }



from django.db import connections, DEFAULT_DB_ALIAS
from django.core.management import color


def sequence_reset_sql(model, using=DEFAULT_DB_ALIAS):
    connection = connections[using]
    return connection.ops.sequence_reset_sql(color.no_style(), [model])[0]


def table_exists(table, schema='public', using=DEFAULT_DB_ALIAS):
    connection = connections[using]
    catalog = connection.settings_dict['NAME']
    cursor = connection.cursor()
    cursor.execute('''
        SELECT (1) FROM "information_schema"."tables"
        WHERE "table_catalog" = %s
            AND "table_schema" = %s
            AND "table_name" = %s
    ''', [catalog, schema, table])
    return True if cursor.fetchone() else False
